(*
    Confluence System Design Language Compiler
    Copyright (C) 2003-2004 Tom Hawkins (tomahawkins@yahoo.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*)


(* Nondeterministic Finite Automata (NFA) *)

(* Types. *)

type state_id     = State of int;;
type character_id = Character of int;;

module OrderedState =
  struct
    type t = state_id
    let compare a b =
      match (a, b) with
        (State a, State b) -> compare a b
  end;;

module StateSet = Set.Make(OrderedState);;

type transition = Trans  of character_id * state_id (* input, next state *)
                | TransE of state_id;;              (* next state *)

type nfa = {
  mutable nfaErrors    : StateSet.t;  (* Set of final states resulting in failure. *)
  nfaTransitions         : (state_id, transition list) Hashtbl.t;
  mutable nfaAlphabet    : (PncAst.expression * character_id) list;
  mutable nfaArity       : int;
  mutable nfaNextStateId : int;
  mutable nfaNextCharId  : int;
};;


(* StateSet helpers. *)

let stateSetInfo states =
  "{" ^ StateSet.fold (fun state sofar -> sofar ^ " " ^ string_of_int (match state with State sid -> sid)) states "" ^ " }";;


(* NFA functions. *)

let newNfa () =
  { nfaErrors    = StateSet.empty;
    nfaTransitions = Hashtbl.create 32;
    nfaAlphabet    = [];
    nfaArity       = 0;
    nfaNextStateId = 0;
    nfaNextCharId  = 0};;

let addState nfa =  (* First added state is initial state. *)
  let id = State nfa.nfaNextStateId in
  nfa.nfaNextStateId <- nfa.nfaNextStateId + 1;
  Hashtbl.add nfa.nfaTransitions id [];
  id;;

let setInvalid nfa state =
  nfa.nfaErrors <- StateSet.add state nfa.nfaErrors;;

let addCharacter nfa expr =
  if List.mem_assoc expr nfa.nfaAlphabet then List.assoc expr nfa.nfaAlphabet else
    let id = Character nfa.nfaNextCharId in
    nfa.nfaArity      <- nfa.nfaArity + 1;
    nfa.nfaNextCharId <- nfa.nfaNextCharId + 1;
    nfa.nfaAlphabet   <- nfa.nfaAlphabet @ [(expr, id)];
    id;;

let addTransition nfa state character nextState =
  Hashtbl.replace nfa.nfaTransitions state (Trans (character, nextState) :: Hashtbl.find nfa.nfaTransitions state);;

let addETransition nfa state nextState =
  Hashtbl.replace nfa.nfaTransitions state (TransE nextState :: Hashtbl.find nfa.nfaTransitions state);;

let getTransitions nfa state =
  Hashtbl.find nfa.nfaTransitions state;;

let getAlphabet nfa =
  nfa.nfaAlphabet;;

let getArity nfa =
  nfa.nfaArity;;

let findNextStates nfa input states =
  let arity = getArity nfa in
  let f2 character =
    match character with
      Character cid ->
        0 <> input land (1 lsl (arity - 1 - cid))
  in
  let f1 states transition =
    match transition with
      Trans (character, state) -> if f2 character then StateSet.add state states else states
    | TransE state -> StateSet.add state states
  in
  let f0 state states =
    List.fold_left f1 states (getTransitions nfa state)
  in
  StateSet.fold f0 states StateSet.empty
;;

let isGoodState nfa states =
  StateSet.is_empty (StateSet.inter nfa.nfaErrors states);;

let nfaInfo nfa =
  let rec infoChars chars =
    match chars with
      [] -> ""
    | (expr, Character id) :: chars -> 
        "  Character" ^ string_of_int id ^ ": " ^ PncAst.infoExpr expr ^ "\n" ^ infoChars chars
  in
  let rec infoTrans trans =
    match trans with
      [] -> ""
    | Trans (Character cid, State sid) :: trans ->
        "    Character" ^ string_of_int cid ^ " -> State" ^ string_of_int sid ^ "\n" ^ infoTrans trans
    | TransE (State sid) :: trans ->
        "    E-Transition -> State" ^ string_of_int sid ^ "\n" ^ infoTrans trans
  in
  let rec infoStates id =
    if id = nfa.nfaNextStateId then "" else
      "  State" ^ string_of_int id ^ ":" ^
      (if StateSet.mem (State id) nfa.nfaErrors then " Final State Invalid" else "") ^ "\n" ^
      infoTrans (getTransitions nfa (State id)) ^ "\n" ^
      infoStates (id + 1)
  in
  "NFA Inputs:\n\n" ^
  infoChars (getAlphabet nfa) ^ "\n" ^
  "NFA State Transitions:\n\n" ^
  infoStates 0 ^ "\n"
;;


(* Building NFA from PNC. *)

let pncToNfa states =
  let nfa = newNfa () in
  let stateToSid = Hashtbl.create 32 in
  let exprToCid = Hashtbl.create 32 in
  let states = match states with PncAst.States states -> states in
  let rec addChar trans =
    match trans with
      [] -> ()
    | PncAst.Trans (expr, _) :: trans ->
        if not (Hashtbl.mem exprToCid expr) then Hashtbl.add exprToCid expr (addCharacter nfa expr);
        addChar trans
    | PncAst.TransE _ :: trans ->
        addChar trans
  in
  let addStateAndChars state =
    match state with
      PncAst.StateValid name ->
        Hashtbl.add stateToSid name (addState nfa)
    | PncAst.StateCond (name, trans) ->
        Hashtbl.add stateToSid name (addState nfa);
        addChar trans
    | PncAst.StateInvalid name ->
        let sid = addState nfa in
        Hashtbl.add stateToSid name sid;
        setInvalid nfa sid
  in
  let rec addTrans sid trans =
    match trans with
      [] -> ()
    | PncAst.Trans (expr, state) :: trans ->
        addTransition nfa sid (Hashtbl.find exprToCid expr) (Hashtbl.find stateToSid state);
        addTrans sid trans;
    | PncAst.TransE state :: trans ->
        addETransition nfa sid (Hashtbl.find stateToSid state);
        addTrans sid trans;
  in
  let addStateTrans state =
    match state with
      PncAst.StateValid _ | PncAst.StateInvalid _ -> ()
    | PncAst.StateCond (state, trans) ->
        addTrans (Hashtbl.find stateToSid state) trans
  in

  List.iter addStateAndChars states;
  List.iter addStateTrans states;
  nfa
;;


(* Converting NFA to CF StateMachine Table. *)

let nfaToTable nfa =
  let nextStateCode = ref 1 in
  let nextTransCode = ref 0 in
  let stateCodes = Hashtbl.create 32 in
  let table = ref [] in
  let queue = Queue.create () in
  let lastInput = (1 lsl (getArity nfa)) - 1 in
  Hashtbl.add stateCodes (StateSet.singleton (State 0)) 0;
  Queue.add (StateSet.singleton (State 0)) queue;
  while not (Queue.is_empty queue) do
    let currentStates = Queue.take queue in
    let currentStateCode = Hashtbl.find stateCodes currentStates in
    for i = 0 to lastInput do
      let nextStates = findNextStates nfa i currentStates in
      if not (Hashtbl.mem stateCodes nextStates) then begin
        Hashtbl.add stateCodes nextStates !nextStateCode;
        incr nextStateCode;
        Queue.add nextStates queue;
      end;
      table := (i, currentStateCode, Hashtbl.find stateCodes nextStates, isGoodState nfa nextStates, (if isGoodState nfa currentStates && isGoodState nfa nextStates then (incr nextTransCode; !nextTransCode - 1) else (-1))) :: !table;
    done;
  done;
  (List.rev !table, !nextTransCode);;

let rec intToBin value width sofar =
  if width <= 0 then sofar else
    intToBin (value / 2) (width - 1) ((if value mod 2 = 1 then "1" else "0") ^ sofar);;

let writeCfStateMachine write states coverage =
  let nfa = pncToNfa states in
  let (table, numOfGoodTransitions) = nfaToTable nfa in
  let inputArity = getArity nfa in
  let encodeInput value =
    intToBin value inputArity ""
  in
  let encodeTransition value =
    if coverage then
      if value <> (-1) then
        String.make (numOfGoodTransitions - 1 - value) '0' ^ "1" ^ String.make value '0'
      else
        String.make numOfGoodTransitions '0'
    else
      ""
  in
  let rec writeExpr expr =
    match expr with
      PncAst.ExprName name ->
        write (PncAst.nameUpperCase name)
    | PncAst.ExprNot expr ->
        write ("('~' ");
        writeExpr expr;
        write (")");
    | PncAst.ExprAnd (exprLeft, exprRight) ->
        write ("(");
        writeExpr exprLeft;
        write (" '&' ");
        writeExpr exprRight;
        write (")");
    | PncAst.ExprOr  (exprLeft, exprRight) ->
        write ("(");
        writeExpr exprLeft;
        write (" '|' ");
        writeExpr exprRight;
        write (")");
  in
  let rec writeInputs inputs =
    match inputs with
      [] -> ()
    | [(expr, _)] ->
        writeExpr expr;
    | (expr, _) :: inputs ->
        writeExpr expr;
        write (" '++' ");
        writeInputs inputs;
  in
  let writeEntry (input, state, nextState, good, transition) =
    write ("    [\"" ^ encodeInput input ^ "\" " ^ string_of_int state ^ " " ^ string_of_int nextState ^ " \"" ^ (if good then "1" else "0") ^ encodeTransition transition ^ "\"]\n")
  in
  write ("{StateMachine, [\n");
  List.iter (fun entry -> writeEntry entry) table;
  write ("  ] (");
  writeInputs (getAlphabet nfa);
  write ("), $}\n");
;;

