(*
    Confluence System Design Language Compiler
    Copyright (C) 2003-2004 Tom Hawkins (tomahawkins@yahoo.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*)


(* Types. *)

type producer = Nls.id;;

type system = System     of system *          (* Parent system. *)
                            int *             (* SysId. *)
                            Nls.space *       (* Nls space. *)
                            bool ref *        (* Cre determined. *)
                            string ref *      (* Clock domain. *)
                            producer ref *    (* Reset. *)
                            producer ref *    (* Enable. *)
                            Loc.loc list      (* Call path. *)
            | SystemRoot of int *             (* SysId. *)
                            Nls.space *       (* Nls space. *)
                            string *          (* Clock domain. *)
                            producer *        (* Reset. *)
                            producer;;        (* Enable. *)

type stateful = Reg     of system * producer * producer * producer (* system, reg, rbuf, ebuf *)
              | Mem     of system * producer * producer            (* system, ram, ebuf *)
              | Ext     of system * producer * producer * producer (* system, ext, rbuf, ebuf *)
;;



(* Local stateful data. *)

let subResets  = Hashtbl.create 64;;    (* SysId to producer IdSet. *)
let subEnables = Hashtbl.create 64;;    (* SysId to producer IdSet. *)
let nextSysId    = ref 0;;              (* Next SysId. *)
let statefuls    = ref [];;             (* list of stateful components. *)
let nls          = Nls.newNls ();;  (* Working nls. *)

let genXml       = ref false;;
let genVerilog   = ref false;;
let genVhdl      = ref false;;
let genC         = ref false;;
let genNuSmv     = ref false;;
let genTestBench = ref (-1);;
let genSignalMap = ref false;;
let fileName     = ref "default_file";;
let compileC     = ref false;;


(* Setting constraints. *)

let setFileName name =
  fileName := name;
  Nls.setFile nls name
;;

let setBuildName name =
  Nls.setBuild nls name;;

let setHeader docs =
  Nls.setHeader nls docs;;

let setGenXml flag =
  genXml := flag;;

let setGenVerilog flag =
  genVerilog := flag;;

let setGenVhdl flag =
  genVhdl := flag;;

let setGenC flag =
  genC := flag;;

let setGenNuSmv flag =
  genNuSmv := flag;;

let setGenTestBench cycles =
  genTestBench := cycles;;

let setGenSignalMap flag =
  genSignalMap := flag;;

let setCompileC flag =
  compileC := flag
;;

let setEmbeddedC flag =
  Lp.setEmbeddedC flag
;;

let setExternalC name =
  Lp.setExternalC name
;;
  

(* System functions. *)

let systemSpace system =
  match system with
    SystemRoot (_, space, _, _, _)
  | System (_, _, space, _, _, _, _, _) -> space;;

let systemId system =
  match system with
    SystemRoot (id, _, _, _, _)
  | System (_, id, _, _, _, _, _, _) -> id;;

let systemLocs system =
  match system with
    SystemRoot _ -> []
  | System (_, _, _, _, _, _, _, locs) -> locs;;

let newRootSys () =
  let id = !nextSysId in
  incr nextSysId;
  let space = Nls.getSpaceMain nls in
  SystemRoot (id, space, "clock", Nls.newVecConst nls space [] "0", Nls.newVecConst nls space [] "1");;  (* "clock" is default. *)

let newSubSys parentSys appLocs compLoc =
  let id = !nextSysId in
  incr nextSysId;
  System (parentSys, id, Nls.newSpaceSub (systemSpace parentSys) (Loc.toString (List.hd appLocs)) (Loc.toString compLoc), ref false, ref "", ref Nls.fakeId, ref Nls.fakeId, appLocs);;
  
let rec determineCre system =  (* Returns (clock, reset producer, enable producer) *)
  (*
    clock  = pClock when clock = "" else clock
    reset  = parentReset OR localResets
    enable = parentEnable AND (parentReset OR localResets) 
  *)
  match system with
    SystemRoot (_, _, clock, reset, enable) -> (clock, reset, enable)
  | System (parent, id, space, creDet, clock, reset, enable, locs) ->
      if not !creDet then begin
        creDet := true;
        let (pClock, pReset, pEnable) = determineCre parent in
        let pid = systemId parent in
        let pSpace = systemSpace parent in
        let foldOr a b =
          Nls.newVecOr nls pSpace locs a b
        in
        let foldAnd a b =
          Nls.newVecAnd nls pSpace locs a b
        in

        if !clock = "" then clock := pClock;

        if not (Hashtbl.mem subResets id) then
          reset := pReset
        else
          reset := Nls.IdSet.fold foldOr (Hashtbl.find subResets id) pReset;

        if not (Hashtbl.mem subEnables id) then
          enable := pEnable
        else
          enable := Nls.IdSet.fold foldAnd (Hashtbl.find subEnables id) pEnable;
      end;
      (!clock, !reset, !enable);;


(* Finding signal width. *)

let producerWidth p = Nls.nodeWidth nls p;;


(* Synchronization functions. *)

let setClockDomain system clockName error =
  match system with
    SystemRoot _ -> error "System error.  Can not set clock domain for root system."
  | System (_, _, _, _, clock, _, _, _) ->
      if !clock <> "" then error ("System error.  System already assigned clock domain \"" ^ !clock ^ "\".  Can not be set to \"" ^ clockName ^ "\".");
      clock := clockName;;

let addSubReset system producer error =
  match system with
    SystemRoot _ -> error "System error.  Can not reset root system."
  | System (_, id, _, _, _, _, _, _) ->
      if Hashtbl.mem subResets id then
        Hashtbl.replace subResets id (Nls.IdSet.add producer (Hashtbl.find subResets id))
      else
        Hashtbl.add subResets id (Nls.IdSet.singleton producer);;

let addSubEnable system producer error  =
  match system with
    SystemRoot _ -> error "System error.  Can not enable root system."
  | System (_, id, _, _, _, _, _, _) ->
      if Hashtbl.mem subEnables id then
        Hashtbl.replace subEnables id (Nls.IdSet.add producer (Hashtbl.find subEnables id))
      else
        Hashtbl.add subEnables id (Nls.IdSet.singleton producer);;


(* Helpers. *)

let bits n =
  Lp.bits n;;

let rec zeros width =
  if width <= 0 then ""
  else
    "0" ^ zeros (width - 1);;


(* Logic primitive constructors.  Given owner space, input producers, and configuration parameters, returns new producer. *)

let newVecInput system name position width =
  Nls.newVecInput nls (systemSpace system) (systemLocs system) name position width;;

let newVecOutput system name position producer =
  Nls.newVecOutput nls (systemSpace system) (systemLocs system) name position (producerWidth producer) producer;;

let newVecInOut system name position ctrlProducer dataProducer =
  Nls.newVecInOut nls (systemSpace system) (systemLocs system) name position (producerWidth dataProducer) ctrlProducer dataProducer;;

let newVecAssert  system test msg = Nls.newVecAssert nls (systemSpace system) (systemLocs system) test msg;;
let newVecPrint   system enable   msg = Nls.newVecPrint  nls (systemSpace system) (systemLocs system) enable msg;;
let newVecConst   system value  = Nls.newVecConst  nls (systemSpace system) (systemLocs system) value;;
let newVecBuf     system p0     = Nls.newVecBuf    nls (systemSpace system) (systemLocs system) p0;;
let newVecNot     system p0     = Nls.newVecNot    nls (systemSpace system) (systemLocs system) p0;;
let newVecAnd     system p0 p1  = Nls.newVecAnd    nls (systemSpace system) (systemLocs system) p0 p1;;
let newVecOr      system p0 p1  = Nls.newVecOr     nls (systemSpace system) (systemLocs system) p0 p1;;
let newVecXor     system p0 p1  = Nls.newVecXor    nls (systemSpace system) (systemLocs system) p0 p1;;
let newVecConcat  system p0 p1  = Nls.newVecConcat nls (systemSpace system) (systemLocs system) p0 p1;;
let newVecSelect  system p0 bs  = Nls.newVecSelect nls (systemSpace system) (systemLocs system) p0 bs;;
let newVecEqu     system p0 p1  = Nls.newVecEqu    nls (systemSpace system) (systemLocs system) p0 p1;;
let newVecAdd     system p0 p1  = Nls.newVecAdd    nls (systemSpace system) (systemLocs system) p0 p1;;
let newVecSub     system p0 p1  = Nls.newVecSub    nls (systemSpace system) (systemLocs system) p0 p1;;
let newVecMul     system p0 p1  = Nls.newVecMul    nls (systemSpace system) (systemLocs system) p0 p1;;
let newVecSwitch  system pred t f = Nls.newVecSwitch nls (systemSpace system) (systemLocs system) pred t f;;


(* External combinatorial and sequential components. *)

let newVecExtComb system name params dataIn widthOut =
  Nls.newVecExtComb nls (systemSpace system) (systemLocs system) name params dataIn widthOut;;

let newVecExtSequ system name params dataIn widthOut =
  let rbuf = newVecBuf system (newVecConst system "0") in
  let ebuf = newVecBuf system (newVecConst system "0") in
  let ext  = Nls.newVecExtSequ nls (systemSpace system) (systemLocs system) name params rbuf ebuf dataIn widthOut in
  statefuls := Ext (system, ext, rbuf, ebuf) :: !statefuls;
  ext;;


(** Stateful logic primitives. *)

let newVecState system width =
  newVecBuf system (newVecConst system (String.make width '0'))
;;

let newVecReg system dataIn dataOut =
  let rbuf = newVecBuf system (newVecConst system "0") in
  let ebuf = newVecBuf system (newVecConst system "0") in
  let reg = Nls.newVecReg nls (systemSpace system) (systemLocs system) rbuf ebuf dataIn in
  Nls.reconnect nls dataOut 0 reg;
  statefuls := Reg (system, reg, rbuf, ebuf) :: !statefuls
;;
    
let newVecRom system width values addr data =
  let ebuf = newVecBuf system (newVecConst system "0") in
  let rom = Nls.newVecRom nls (systemSpace system) (systemLocs system) ebuf width values addr in
  Nls.reconnect nls data 0 rom;
  statefuls := Mem (system, rom, ebuf) :: !statefuls
;;

let newVecRamMem system width values =
  Nls.newVecRamMem nls (systemSpace system) (systemLocs system) width values;;

let newVecRamWrite system mem addr data =
  let ebuf = newVecBuf system (newVecConst system "0") in
  let ram  = Nls.newVecRamWrite nls (systemSpace system) (systemLocs system) mem ebuf addr data in
  statefuls := Mem (system, ram, ebuf) :: !statefuls
;;

let newVecRamRead system mem rbw addr data =
  let ebuf = newVecBuf system (newVecConst system "0") in
  let ram  = Nls.newVecRamRead nls (systemSpace system) (systemLocs system) rbw mem ebuf addr in
  Nls.reconnect nls data 0 ram;
  statefuls := Mem (system, ram, ebuf) :: !statefuls
;;


(* Post evaluation functions. *)

let rec connectStateful stateful =
  match stateful with
    Reg (system, node, rbuf, ebuf) | Ext (system, node, rbuf, ebuf)  ->
      let (clock, reset, enable) = determineCre system in
      Nls.reconnect nls rbuf 0 reset;
      Nls.reconnect nls ebuf 0 enable;
      Nls.setClock  nls node clock

  | Mem (system, node, ebuf) ->
      let (clock, _, enable) = determineCre system in
      Nls.reconnect nls ebuf 0 enable;
      Nls.setClock  nls node clock;
;;
      
let checkValidTestBench () =
  if not (Nls.isValidTestBench nls) then Report.fatal "** Code Gen Error\n\n  Invalid testbench.  Testbench must have a single clock, 1 or more outputs,  and no inputs or inouts.";;

let checkValidDesign () =
  if not (Nls.isGeneratable nls) then Report.fatal "** Code Gen Error\n\n  Invalid design.  Design must have outputs, inouts, or assertions.";;

let genCode () =
  if !genVerilog || !genVhdl || !genC || !genNuSmv || !genXml then begin
    checkValidDesign ();
    List.iter connectStateful !statefuls;
    NlsOpt.constantPropagation nls;
    NlsOpt.removeNullEffect nls;
    NlsOpt.checkConnections nls;

    NlsSpace.populateSpaces nls;
    if !genXml       then NlsSpace.codeXmlNetlist nls;
    if !genSignalMap then NlsSpace.codeSignalMap  nls;

    if !genVerilog          then NlsVerilog.code nls;
    if !genVhdl             then NlsVhdl.code nls;
    if !genC                then (NlsC.code nls; NlsC.codeSimulator nls);

    if !genNuSmv            then NlsNuSmv.code nls;
    if !genVerilog  && !genTestBench >= 0 then begin checkValidTestBench (); NlsVerilog.codeTestBench  nls !genTestBench; end;
    if !genVhdl     && !genTestBench >= 0 then begin checkValidTestBench (); NlsVhdl.codeTestBench     nls !genTestBench; end;
    if !genC        && !genTestBench >= 0 then begin checkValidTestBench (); NlsC.codeTestBench        nls !genTestBench; end;

    if !genC && !compileC then
      let cmd = "gcc -Wall -O2 -o " ^ !fileName ^ " " ^ !fileName ^ "_sim.c " ^ !fileName ^ ".c" in
      let _ = Sys.command cmd in
      ();

  end;;


