(*
    Confluence System Design Language Compiler
    Copyright (C) 2003-2004 Tom Hawkins (tomahawkins@yahoo.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*)


type docTree = Empty
             | Contents of string * (string * docTree) list * (string * fileInfo) list
and fileInfo = string * string * string list * string list * compDoc list
             (* doc, env, exports, imports, compDocs *)
and compDoc  = string * string * string list
;;

let nextId = ref 0;;
let ids = Hashtbl.create 64;;


(** Functions for building doc tree. *)

let isDir file =
  try
    let _ = Sys.readdir file in
    true
  with Sys_error _ ->
    false
;;

let isCfFile file =
  let len = String.length file in
  not (isDir file) && len >= 3 && String.sub file (len - 3) 3 = ".cf"
;;

let readAll file =
  try
    let file = open_in file in
    let docs = ref "" in
    try
      while true do
        docs := !docs ^ input_line file ^ "\n"
      done;
      !docs
    with End_of_file ->
      close_in file;
      !docs
  with Sys_error _ ->
    ""
;;

let rec exploreDir dirPath =
  if not (isDir dirPath) then Report.fatal ("** Error:  Unknown directory: " ^ dirPath);
  Hashtbl.add ids dirPath !nextId; incr nextId;
  match exploreContents dirPath (Array.to_list (Sys.readdir dirPath)) with
  | ("", [], []) -> Empty
  | (doc, dirInfo, fileInfo) -> Contents (doc, List.sort (fun (a, _) (b, _) -> compare a b) dirInfo, List.sort (fun (a, _) (b, _) -> compare a b) fileInfo)

and exploreContents dirPath files =
  match files with
  | [] -> ("", [], [])
  | file :: files ->
    let (doc, dirInfo, fileInfo) = exploreContents dirPath files in
    let filePath = dirPath ^ file in
    if file = "__cfdoc__" then
      (readAll filePath, dirInfo, fileInfo)
    else if isDir filePath then
      match exploreDir (filePath ^ "/") with
      | Empty -> (doc, dirInfo, fileInfo)
      | Contents _ as dirInfo0 -> (doc, (file, dirInfo0) :: dirInfo, fileInfo)
    else if isCfFile filePath then begin
      Hashtbl.add ids filePath !nextId; incr nextId;
      (doc, dirInfo, (file, CfParserUtil.parse_doc filePath CfLexer.token CfParser.file) :: fileInfo)
    end else
      (doc, dirInfo, fileInfo)
;;


(** Functions to print the doc tree. *)

let fileOfPath path =
  let n = Hashtbl.find ids path in
  "n" ^ string_of_int n ^ ".html"
;;

let rec pathLinks paths =
  match paths with
  | [] -> "<a href=\"index.html\">Index</a> : "
  | (name, path) :: paths ->
      pathLinks paths ^ "<a href=\"" ^  fileOfPath path ^ "\">" ^ name ^ "</a> / "
;;

(*
(** Split path into a list of strings; root at bottom. *)
let rec splitPath path =
  match path with
  | "" -> []
  | "/" -> ["/"]
  | path ->
      if String2.right_char path = '/' then
        splitPath (String2.left_string path)
      else
        try
          let i = String.rindex path '/' in
          String.sub path (i + 1) (String.length path - i - 1) :: splitPath (String.sub path 0 (i + 1))
        with Not_found ->
          [path]
;;

(** Clean path.  Remove redundent '//'. *)
let cleanPath path =
  let rec f paths =
    match paths with
    | [] -> ""
    | ["/"] -> ""
    | [path] -> path
    | path :: paths -> f paths ^ "/" ^ path
  in
  f (splitPath path)
;;
*)

let baseEnv =
  try
    let base = String2.replace_xml_chars (Sys.getenv "CF_ENV") in
    "<a href=\"" ^ base ^ "\">" ^ base ^ "</a>"
  with Not_found ->
    "Unknown Base Environment (CF_ENV)"
;;

let insertMetaChars s =
  String2.insert_meta_chars s ['\\', "\\"; ')', ")"; '*', "*"]
;;

let writeFile htmlDir filePath parentPaths file (docs, env, exports, imports, compDocs) =
  let f = open_out (htmlDir ^ "/" ^ fileOfPath filePath) in

  output_string f ("<html><head><title>Confluence Source File: " ^ file ^ "</title></head><body>\n");
  output_string f ("<h2>");
  output_string f (pathLinks parentPaths ^ file); 
  output_string f ("</h2>\n");

  output_string f ("<h3>Source: <a href=\"../../" ^ filePath ^ "\">" ^ file ^ "</a></h3>\n");

  if docs <> "" then begin
    output_string f ("<h3>Documentation</h3>\n");
    output_string f ("<pre>" ^ String2.replace_xml_chars (insertMetaChars docs) ^ "</pre>\n");
  end;

  let env = String2.replace_xml_chars env in
  output_string f ("<h3>Environment: " ^ (if env = "" then baseEnv else if env = "root" then "Root Environment" else "<a href=\"" ^ env ^ "\">" ^ env ^ "</a>") ^ "</h3>\n"); (* XXX env is not a correct path. *)

  if exports <> [] then begin
    output_string f ("<h3>Exports</h3>\n");
    output_string f ("<ul>\n");
    List.iter (fun export -> output_string f ("<li>" ^ String2.replace_xml_chars export ^ "</li>\n")) exports;
    output_string f ("</ul>\n");
  end;

  if imports <> [] then begin
    output_string f ("<h3>Imports</h3>\n");
    output_string f ("<ul>\n");
    List.iter (fun import -> output_string f ("<li>" ^ String2.replace_xml_chars import ^ "</li>\n")) imports;
    output_string f ("</ul>\n");
  end;

  if compDocs <> [] then begin
    output_string f ("<h3>Components</h3>\n");
    List.iter (fun (doc, comp, ports) ->
      output_string f ("<h4>{" ^ comp);
      List.iter (fun port -> output_string f (" " ^ port)) ports;
      output_string f ("}</h4>\n");
      output_string f ("<pre>" ^ String2.replace_xml_chars (insertMetaChars doc) ^ "</pre>\n")
    ) (List.rev compDocs);
  end;

  output_string f ("</body></html>\n");
  close_out f
;;

(* Variable name convensions:

     dir  : Directory name.  No '/'
     file : File name.  Could be a dir.  No '/'.
     dirPath : Path name with '/'.  Ends in '/'.
     filePath : Path name ending in .cf.
*)


let rec writeDir html dir dirPath dirs docTree =
  let parentPaths = (dir, dirPath) :: dirs in
  match docTree with
  | Empty -> ()
  | Contents (doc, subDirs, subFiles) ->
      let f = open_out (html ^ fileOfPath dirPath) in

      output_string f ("<html><head><title>Confluence Source Directory: " ^ dir ^ "</title></head><body>\n");
      output_string f ("<h2>");
      output_string f (pathLinks dirs ^ dir); 
      output_string f ("</h2>\n");

      if doc <> "" then begin
        output_string f ("<h3>Documentation</h3>\n");
        output_string f ("<pre>" ^ doc ^ "</pre>\n");
      end;

      if subDirs <> [] then begin
        output_string f ("<h3>Directories</h3>\n");
        output_string f ("<ul>\n");
        List.iter (fun (subDir, subDirTree) ->
          let subDirPath = dirPath ^ subDir ^ "/" in
          output_string f ("<li><a href=\"" ^ fileOfPath subDirPath ^ "\">" ^ subDir ^ "</a></li>\n");
          writeDir html subDir subDirPath parentPaths subDirTree
        ) subDirs;
        output_string f ("</ul>\n");
      end;

      if subFiles <> [] then begin
        output_string f ("<h3>Files</h3>\n");
        output_string f ("<ul>\n");
        List.iter (fun (file, fileInfo) ->
          let filePath = dirPath ^ file in
          output_string f ("<li><a href=\"" ^ fileOfPath filePath ^ "\">" ^ file ^ "</a></li>\n");
          writeFile html filePath parentPaths file fileInfo
        ) subFiles;
        output_string f ("</ul>\n");
      end;

      output_string f ("</body></html>\n");
      close_out f
;;

let writeIndex html dir docTree =
  let f = open_out (html ^ "index.html") in
  let rec writeDir dirPath docTree =
    match docTree with
    | Empty -> ()
    | Contents (_, dirs, files) ->
        output_string f ("<ul>\n");
        List.iter (fun (dir, docTree) ->
          output_string f ("<li><a href=\"" ^ fileOfPath (dirPath ^ dir ^ "/") ^ "\">" ^ dir ^ "</a></li>\n");
          writeDir (dirPath ^ dir ^ "/") docTree
        ) dirs;
        List.iter (fun (file, _) ->
          output_string f ("<li><a href=\"" ^ fileOfPath (dirPath ^ file) ^ "\">" ^ file ^ "</a></li>\n")
        ) files;
        output_string f ("</ul>\n")
  in
    output_string f ("<html><head><title>Confluence Source Directory Index</title></head><body>\n");
    output_string f ("<h2>Source Directory Index</h2>\n");
    output_string f ("<ul><li><a href=\"" ^ fileOfPath (dir ^ "/") ^ "\">" ^ dir ^ "</a></li>\n");
    writeDir (dir ^ "/") docTree;
    output_string f ("</ul>\n");
    output_string f ("</body></html>\n");
    close_out f
;;


let hasSlash name =
  try
    let i = String.index name '/' in
    true
  with Not_found ->
    false
;;


(** Show help. *)
let info () =
    print_string "\n";
    print_string ("Confluence Documentation Generation " ^ Version.version ^ "\n\n");
    print_string "Copyright (C) 2003-2004 Tom Hawkins\n\n";
    print_string "Usage:  cfdoc [-h] [-i src_directory]\n\n";
    exit 0
;;

(** Main. *)
let main () =

  let srcDir = ref "src" in
  let i = ref 1 in
  while !i < Array.length Sys.argv do
    match Sys.argv.(!i) with
    | "-i" ->
        srcDir := Sys.argv.(!i + 1);
        i := !i + 2
    | _ ->
        info ()
  done;
  let srcDir = !srcDir in
  let docDir = srcDir ^ "/doc/" in

  if hasSlash srcDir then Report.fatal ("** Error:  cfdev must be called above source directory.");
  if not (Sys.file_exists srcDir && isDir srcDir) then Report.fatal ("** Error:  Source directory does not exist:  " ^ srcDir);
  if not (Sys.file_exists docDir && isDir docDir) then Report.fatal ("** Error:  Documentation directory does not exist:  " ^ docDir);

  let docTree = exploreDir (srcDir ^ "/") in
  writeIndex docDir srcDir docTree;
  writeDir docDir srcDir (srcDir ^ "/") [] docTree;

  exit 0
;;

main();;

