(*
    Confluence System Design Language Compiler
    Copyright (C) 2003-2004 Tom Hawkins (tomahawkins@yahoo.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*)


(* Exceptions. *)

exception Error of string;;


(* Logic primitive types. *)

type lp =
| VecInput    of string * int * int            (* Params: name, position, width.    Inputs: none.                                       Output: input. *)
| VecOutput   of string * int * int            (* Params: name, position, width.    Inputs: output.                                     Output: none. *)
| VecInOut    of string * int * int            (* Params: name, position, width.    Inputs: ctrl (drive output when high), data_out.    Output: data_in. *)
| VecAssert   of string                        (* Params: message.                  Inputs: assertion (message when low).               Output: none. *)
| VecPrint    of int                           (* Params: width of print vector.    Inputs: enable, vector.                             Output: none. *)
| VecConst    of string                        (* Params: binary value.             Inputs: none.                                       Output: constant. *)
| VecBuf      of int                           (* Params: width. *)
| VecNot      of int                           (* Params: width. *)
| VecAnd      of int                           (* Params: width. *)
| VecOr       of int                           (* Params: width. *)
| VecXor      of int                           (* Params: width. *)
| VecConcat   of int * int                     (* Params: width_left, width_right.  Inputs: left, right.                                Output: concatenation. *)
| VecSelect   of int * int list                (* Params: width, bits selected.     Inputs: input.                                      Output: output. *)
| VecEqu      of int                           (* Params: width.                    Inputs: left, right.                                Output: equal. *)
| VecAdd      of int                           (* Params: width.                    Inputs: left, right.                                Output: left + right. *)
| VecSub      of int                           (* Params: width.                    Inputs: left, right.                                Output: left - right. *)
| VecMul      of int                           (* Params: width.                    Inputs: left, right.                                Output: left * right. *)
| VecSwitch   of int                           (* Params: width.                           Inputs: predicate, true, false.                     Output: selected data. *)
| VecExtComb  of string * int list * int * int (* Params: comp name, config params, input width, output width.  Inputs: input.                 Outputs: output. *)
| VecExtSequ  of string * string * int list * int * int (* Params: clock, comp name, config params, input width, output width.  Inputs: reset, enable, input.  Outputs: output. *)
| VecReg      of string * int                  (* Params: clock, width.             Inputs: reset, enable, signal.                      Output: register signal. *)
| VecRom      of string * int * Intbig.intbig list       (* Params: clock, width, values.            Inputs: enable, addr.                               Output: value. *)
| VecRamMem   of int * Intbig.intbig list      (* Params: data_width, initial values.    Inputs: none.   Outputs: memory_link (not a real signal). *)     
| VecRamWrite of string * int * int            (* Params: clock, addr_width, data_width.          Inputs: memory_link, enable, addr, data.  Output: none. *)
| VecRamRead  of string * bool * int * int     (* Params: clock, is_rbw, addr_width, data_width.  Inputs: memory_link, enable, addr.        Output: data. *)
;;


(* Misc. *)

let bits n =
  match n with
    0 -> 0
  | 1 -> 1
  | n when n > 1 -> int_of_float (ceil (log (float_of_int n) /. log 2.))
  | _ -> raise (Error "Function 'bits' requires n greater than zero");;

let rec listCreate count value =
  if count <= 0 then [] else value :: listCreate (count - 1) value;;
  

(* Primitive info. *)

let primString lp =
  match lp with
  | VecInput    _ -> "VecInput"
  | VecOutput   _ -> "VecOutput"
  | VecInOut    _ -> "VecInOut"
  | VecAssert   _ -> "VecAssert"
  | VecPrint    _ -> "VecPrint"
  | VecConst    s -> "VecConst: " ^ s
  | VecBuf      _ -> "VecBuf"
  | VecNot      _ -> "VecNot"
  | VecAnd      _ -> "VecAnd"
  | VecOr       _ -> "VecOr"
  | VecXor      _ -> "VecXor"
  | VecConcat   _ -> "VecConcat"
  | VecSelect   _ -> "VecSelect"
  | VecEqu      _ -> "VecEqu"
  | VecAdd      _ -> "VecAdd"
  | VecSub      _ -> "VecSub"
  | VecMul      _ -> "VecMul"
  | VecSwitch   _ -> "VecSwitch"
  | VecExtComb  _ -> "VecExtComb"
  | VecExtSequ  _ -> "VecExtSequ"
  | VecReg      _ -> "VecReg"
  | VecRom      _ -> "VecRom"
  | VecRamMem   _ -> "VecRamMem"
  | VecRamWrite _ -> "VecRamWrite"
  | VecRamRead  _ -> "VecRamRead"
;;

let arity lp =
  match lp with
    VecInput    _ -> 0
  | VecOutput   _ -> 1
  | VecInOut    _ -> 2
  | VecAssert   _ -> 1
  | VecPrint    _ -> 2
  | VecConst    _ -> 0
  | VecBuf      _ -> 1
  | VecNot      _ -> 1
  | VecAnd      _ -> 2
  | VecOr       _ -> 2
  | VecXor      _ -> 2
  | VecConcat   _ -> 2
  | VecSelect   _ -> 1
  | VecEqu      _ -> 2  
  | VecAdd      _ -> 2
  | VecSub      _ -> 2
  | VecMul      _ -> 2
  | VecSwitch   _ -> 3
  | VecExtComb  _ -> 1
  | VecExtSequ  _ -> 3
  | VecReg      _ -> 3
  | VecRom      _ -> 2
  | VecRamMem   _ -> 0
  | VecRamWrite _ -> 4
  | VecRamRead  _ -> 3
;;

let width lp =
  match lp with
    VecInput    (_, _, w)    -> w
  | VecOutput   (_, _, w)    -> w  (* Returns output width for convienence. *)
  | VecInOut    (_, _, w)    -> w
  | VecAssert   _            -> 0
  | VecPrint    _            -> 0
  | VecConst    s            -> String.length s
  | VecBuf      w            -> w
  | VecNot      w            -> w
  | VecAnd      w            -> w
  | VecOr       w            -> w
  | VecXor      w            -> w
  | VecConcat   (w1, w2)     -> w1 + w2
  | VecSelect   (_, bs)      -> List.length bs
  | VecEqu      w            -> 1
  | VecAdd      w            -> w
  | VecSub      w            -> w
  | VecMul      w            -> w
  | VecSwitch   w            -> w
  | VecExtComb  (_, _, _, w) -> w
  | VecExtSequ  (_, _, _, _, w) -> w
  | VecReg      (_, w)       -> w
  | VecRom      (_, w, _)    -> w
  | VecRamMem   (w, v)       -> w * bits (List.length v)  (* Virtual link. *)
  | VecRamWrite (_, _, w)    -> 0
  | VecRamRead  (_, _, _, w) -> w
;;

let inputWidths lp =
  match lp with
    VecInput    _            -> []
  | VecOutput   (_, _, w)    -> w :: []
  | VecInOut    (_, _, w)    -> 1 :: w :: []
  | VecAssert   _            -> 1 :: []
  | VecPrint    w            -> 1 :: w :: []
  | VecConst    _            -> []
  | VecBuf      w            -> w :: []
  | VecNot      w            -> w :: []
  | VecAnd      w            -> w :: w :: []
  | VecOr       w            -> w :: w :: []
  | VecXor      w            -> w :: w :: []
  | VecConcat   (w1, w2)     -> w1 :: w2 :: []
  | VecSelect   (w, _)       -> w :: []
  | VecEqu      w            -> w :: w :: []
  | VecAdd      w            -> w :: w :: []
  | VecSub      w            -> w :: w :: []
  | VecMul      w            -> w :: w :: []
  | VecSwitch   w            -> 1 :: w :: w :: []
  | VecExtComb  (_, _, w, _) -> w :: []
  | VecExtSequ  (_, _, _, w, _) -> 1 :: 1 :: w :: []
  | VecReg      (_, w)       -> 1 :: 1 :: w :: []
  | VecRom      (_, _, v)    -> 1 :: bits (List.length v) :: []
  | VecRamMem   _            -> []
  | VecRamWrite (_, aw, dw)    -> aw * dw :: 1 :: aw :: dw :: []
  | VecRamRead  (_, _, aw, dw) -> aw * dw :: 1 :: aw :: []
;;

let clockDomain lp =
  match lp with
  | VecExtSequ  (c, _, _, _, _)
  | VecReg      (c, _)   
  | VecRom      (c, _, _)
  | VecRamWrite (c, _, _)
  | VecRamRead  (c, _, _, _) -> c
  | _ -> raise (Error "Lp.clockDomain: Combinatorial logic primitive does not below to a clock domain.");;
    



(* Global settings and string writer. *)

let writer = ref (function (str : string) -> ());;

let settingEmbeddedC = ref false;;

let settingExternalC = ref "";;

let setWriter newWriter =
  writer := newWriter;;

let setEmbeddedC flag =
  settingEmbeddedC := flag;;

let setExternalC str =
  settingExternalC := !settingExternalC ^ str;;
  
(* Writer and setting quering. *)

let write str =
  !writer str;;

let embeddedC () =
  !settingEmbeddedC;;

let externalC () =
  !settingExternalC;;

(* Misc HDL gen. *)

let bytes width =
  (width - 1) / 8 + 1;;

let rec intToBin value width =
  if width <= 0 then "" else
    if (value mod 2) = 1 then
      intToBin (value / 2) (width - 1) ^ "1"
    else
      intToBin (value / 2) (width - 1) ^ "0";;

let isCoded lp =
  match lp with
  | VecAssert _ -> true
  | VecPrint  _ -> true
  | VecRamWrite _ -> true
  | _ -> width lp > 0;;

let argList elements =
  let rec f elements =
    match elements with
      [] -> ()
    | [element] -> write element
    | element :: elements -> write (element ^ ", "); f elements
  in
    write "("; f elements; write ")";;

let rec formatMsg msg =
  if msg = "" then "" else
    (match msg.[0] with
      '\\' -> "\\"
    | '"'  -> "\\\""
    | '\n' -> "\\n    "
    | '\t' -> "\\t"
    | '\r' -> ""
    | c    -> String.make 1 c
    ) ^ formatMsg (String.sub msg 1 (String.length msg - 1));;



