(*
    Confluence System Design Language Compiler
    Copyright (C) 2003-2004 Tom Hawkins (tomahawkins@yahoo.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*)


(* Exceptions. *)

exception Error of string;;


(* Logic primitive types. *)

type lp =
| VecInput    of string * int * int            (* Params: name, position, width.    Inputs: none.                                       Output: input. *)
| VecOutput   of string * int * int            (* Params: name, position, width.    Inputs: output.                                     Output: none. *)
| VecInOut    of string * int * int            (* Params: name, position, width.    Inputs: ctrl (drive output when high), data_out.    Output: data_in. *)
| VecAssert   of string                        (* Params: message.                  Inputs: assertion (message when low).               Output: none. *)
| VecPrint    of int                           (* Params: width of print vector.    Inputs: enable, vector.                             Output: none. *)
| VecConst    of string                        (* Params: binary value.             Inputs: none.                                       Output: constant. *)
| VecBuf      of int                           (* Params: width. *)
| VecNot      of int                           (* Params: width. *)
| VecAnd      of int                           (* Params: width. *)
| VecOr       of int                           (* Params: width. *)
| VecXor      of int                           (* Params: width. *)
| VecConcat   of int * int                     (* Params: width_left, width_right.  Inputs: left, right.                                Output: concatenation. *)
| VecSelect   of int * int list                (* Params: width, bits selected.     Inputs: input.                                      Output: output. *)
| VecEqu      of int                           (* Params: width.                    Inputs: left, right.                                Output: equal. *)
| VecAdd      of int                           (* Params: width.                    Inputs: left, right.                                Output: left + right. *)
| VecSub      of int                           (* Params: width.                    Inputs: left, right.                                Output: left - right. *)
| VecMul      of int                           (* Params: width.                    Inputs: left, right.                                Output: left * right. *)
| VecSwitch   of int                           (* Params: width.                           Inputs: predicate, true, false.                     Output: selected data. *)
| VecExtComb  of string * int list * int * int (* Params: comp name, config params, input width, output width.  Inputs: input.                 Outputs: output. *)
| VecExtSequ  of string * string * int list * int * int (* Params: clock, comp name, config params, input width, output width.  Inputs: reset, enable, input.  Outputs: output. *)
| VecReg      of string * int                  (* Params: clock, width.             Inputs: reset, enable, signal.                      Output: register signal. *)
| VecRom      of string * int * Intbig.intbig list       (* Params: clock, width, values.            Inputs: enable, addr.                               Output: value. *)
| VecRamMem   of int * Intbig.intbig list      (* Params: data_width, initial values.    Inputs: none.   Outputs: memory_link (not a real signal). *)     
| VecRamWrite of string * int * int            (* Params: clock, addr_width, data_width.          Inputs: memory_link, enable, addr, data.  Output: none. *)
| VecRamRead  of string * bool * int * int     (* Params: clock, is_rbw, addr_width, data_width.  Inputs: memory_link, enable, addr.        Output: data. *)
;;


(* Misc. *)

val bits : int -> int;;


(* Primitive info. *)

val primString  : lp -> string;;
val arity       : lp -> int;;
val width       : lp -> int;;
val inputWidths : lp -> int list;;
val clockDomain : lp -> string;;


(* Global settings. *)

val setWriter         : (string -> unit) -> unit;;
val setEmbeddedC      : bool -> unit;;
val setExternalC      : string -> unit;;

val write             : string -> unit;;
val embeddedC         : unit -> bool;;
val externalC         : unit -> string;;


(* Misc functions for HDL code generators. *)

val bytes           : int -> int;;
val intToBin        : int -> int -> string;;
val isCoded         : lp -> bool;;
val argList         : string list -> unit;;
val formatMsg       : string -> string;;


