(*
    Confluence System Design Language Compiler
    Copyright (C) 2003-2004 Tom Hawkins (tomahawkins@yahoo.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*)


(* Misc helpers. *)

let zeros width =
  let rec f width =
    if width <= 0 then "" else "0" ^ f (width - 1)
  in
    string_of_int width ^ "'b" ^ f width;;

let zzzs width =
  let rec f width =
    if width <= 0 then "" else "z" ^ f (width - 1)
  in
    string_of_int width ^ "'b" ^ f width;;

let write = Lp.write;;

let widthDecl width =
  if width = 0 then raise (Invalid_argument "Lp.widthDecl:  Width must be greater than 0.");
  if width = 1 then "" else "[" ^ string_of_int (width - 1) ^ ":0] ";;

let rec codeParams params =
  match params with
    [] -> ""
  | param :: params -> ", " ^ string_of_int param ^ codeParams params;;


(* Declaration coder. *)

let decl lp name =
  if Lp.isCoded lp then
    match (lp, Lp.width lp) with

      | (Lp.VecInput _, w)  ->
          write ("input   " ^ widthDecl w ^ name ^ ";\n");

      | (Lp.VecOutput _, w) ->
          write ("output  " ^ widthDecl w ^ name ^ ";\n");

      | (Lp.VecInOut  _, w) ->
          write ("inout   " ^ widthDecl w ^ name ^ ";\n");
      
      | (Lp.VecAssert _, _) | (Lp.VecPrint _, _) -> ()

      | (Lp.VecReg _, w) | (Lp.VecRom _, w) ->
          write ("reg     " ^ widthDecl w ^ name ^ ";\n")

      | (Lp.VecRamMem (dw, vs), _) ->
          write ("reg     " ^ widthDecl dw ^ name ^ " [" ^ string_of_int (List.length vs - 1) ^ ":0];\n");

      | (Lp.VecRamRead (c, rbw, aw, dw), _) ->
          if rbw then begin
            write ("reg     " ^ widthDecl dw ^ name ^ ";\n");
          end else begin
            write ("wire    " ^ widthDecl dw ^ name ^ ";\n");
            write ("reg     " ^ widthDecl aw ^ name ^ "a" ^ ";\n");
          end

      | (Lp.VecRamWrite _, _) -> ()

      | (_, w) ->
          write ("wire    " ^ widthDecl w ^ name ^ ";\n");
;;


(* Function coder. *)
                    
let func lp name inputs =

  if Lp.arity lp <> List.length inputs then raise (Lp.Error ("LpVerilog.func: Arity does not match input list length.  Primitive: " ^ Lp.primString lp ^ "  Inputs: " ^ string_of_int (List.length inputs)));

  let input n =
    List.nth inputs n
  in

  if Lp.isCoded lp then
    match lp with 
      Lp.VecInput _ -> ()
  
    | Lp.VecOutput (n, _, w) ->
        write ("assign " ^ n ^ "_o = " ^ input 0 ^ ";\n")
  
    | Lp.VecInOut (n, _, w) ->
        write ("assign " ^ n ^ "_b = " ^ input 0 ^ " == 1'b1 ? " ^ input 1 ^ " : " ^ zzzs w ^ ";\n")
  
    | Lp.VecAssert msg ->
        let f () =
          write ("  if (" ^ input 0 ^ " == 1'b0)\n");
          write ("    $display(\"**  Assertion Violation\\n\\n    " ^ Lp.formatMsg msg ^ "\\n\\n\");\n");
        in
        write ("initial\n");
        f ();
        write ("always @ (" ^ input 0 ^ ")\n");
        f ();

    | Lp.VecPrint w ->
        let rec f2 name bh bl =
          if bl >= 0 then begin
            write (", " ^ name ^ "[" ^ string_of_int bh ^ ":" ^ string_of_int bl ^ "]");
            f2 name (bl - 1) (bl - 8);
          end
        in
        let f () =
          write ("  if (" ^ input 0 ^ ")\n");
          write ("    $display(\"");
          for i = 1 to Lp.bytes w do write "%c" done;
          write ("\"");
          f2 (input 1) (w - 1) ((Lp.bytes w - 1) * 8);
          write (");\n");
        in
        write ("initial\n");
        f ();
        write ("always @ (" ^ input 0 ^ " or " ^ input 1 ^ ")\n");
        f ();
  
    | Lp.VecConst s ->
        write ("assign " ^ name ^ " = " ^ string_of_int (String.length s) ^ "'b" ^ s ^ ";\n")
        
    | Lp.VecBuf _ ->
        write ("assign " ^ name ^ " = " ^ input 0 ^ ";\n")
  
    | Lp.VecNot _ ->
        write ("assign " ^ name ^ " = ~" ^ input 0 ^ ";\n")
  
    | Lp.VecAnd _ ->
        write ("assign " ^ name ^ " = " ^ input 0 ^ " & " ^ input 1 ^ ";\n")
  
    | Lp.VecOr _ ->
        write ("assign " ^ name ^ " = " ^ input 0 ^ " | " ^ input 1 ^ ";\n")
  
    | Lp.VecXor _ ->
        write ("assign " ^ name ^ " = " ^ input 0 ^ " ^ " ^ input 1 ^ ";\n")
  
    | Lp.VecConcat (0, 0) -> ()
    | Lp.VecConcat (_, 0) ->
        write ("assign " ^ name ^ " = " ^ input 0 ^ ";\n")
    | Lp.VecConcat (0, _) ->
        write ("assign " ^ name ^ " = " ^ input 1 ^ ";\n")
    | Lp.VecConcat (_, _) ->
        write ("assign " ^ name ^ " = {" ^ input 0 ^ ", "  ^ input 1 ^ "};\n")

    | Lp.VecSelect (w, bs) ->
        let rec select1 name scalar bits =
          match bits with
            [] -> ()
          | bit :: bits ->
              write (",\n  " ^ name ^ (if scalar then "" else "[" ^ string_of_int bit ^ "]"));
              select1 name scalar bits
        in

        let b = string_of_int (List.hd bs)
        and t = List.tl bs
        and scalar = w = 1
        in

        if List.length bs = 1 then
          write ("assign " ^ name ^ " = " ^ input 0 ^ (if scalar then "" else "[" ^ b ^ "]") ^ ";\n")
        else begin
          write ("assign " ^ name ^ " = {" ^ input 0 ^ (if scalar then "" else "[" ^ b ^ "]"));
          select1 (input 0) scalar t;
          write ("};\n");
        end;
  
    | Lp.VecEqu _ ->
        write ("assign " ^ name ^ " = " ^ input 0 ^ " == " ^ input 1 ^ ";\n")
  
    | Lp.VecAdd _ ->
        write ("assign " ^ name ^ " = " ^ input 0 ^ " + " ^ input 1 ^ ";\n")
  
    | Lp.VecSub _ ->
        write ("assign " ^ name ^ " = " ^ input 0 ^ " - " ^ input 1 ^ ";\n")
  
    | Lp.VecMul _ ->
        write ("assign " ^ name ^ " = " ^ input 0 ^ " * " ^ input 1 ^ ";\n")
  
    | Lp.VecSwitch w ->
        write ("assign " ^ name ^ " = " ^ input 0 ^ " ? " ^ input 1 ^ " : " ^ input 2 ^ ";\n");

    | Lp.VecExtComb (comp, params, wi, wo) ->
        write (comp ^ " #(" ^ string_of_int wi ^ ", " ^ string_of_int wo ^ codeParams params ^ ") " ^ name ^ "e (" ^ input 0 ^ ", " ^ name ^ ");\n");

    | Lp.VecExtSequ (cd, comp, params, wi, wo) ->
        write (comp ^ " #(" ^ string_of_int wi ^ ", " ^ string_of_int wo ^ codeParams params ^ ") " ^ name ^ "e (" ^ cd ^ "_c, " ^ input 0 ^ ", " ^ input 1 ^ ", " ^ input 2 ^ ", " ^ name ^ ");\n");

    | Lp.VecReg (cd, w) ->
        write ("initial " ^ name ^ " = " ^ zeros w ^ ";\n");
        write ("always @ (posedge " ^ cd ^ "_c)\n");
        write ("  if (" ^ input 0 ^ " == 1'b1)\n");
        write ("    " ^ name ^ " <= " ^ zeros w ^ ";\n");
        write ("  else if (" ^ input 1 ^ " == 1'b1)\n");
        write ("    " ^ name ^ " <= " ^ input 2 ^ ";\n");
  
    | Lp.VecRom (cd, dw, vs) ->

        let rec rom1 indent addrWidth dataWidth dataName addr values =
          match values with
            [] -> ()
          | value :: values -> 
              write (indent ^ string_of_int addrWidth ^ "'b" ^ Lp.intToBin addr addrWidth ^ " : " ^ dataName ^ " <= " ^ string_of_int dataWidth ^ "'b" ^ Intbig.binary_string_of_intbig value (Intbig.intbig_of_int dw) ^ ";\n");
              rom1 indent addrWidth dataWidth dataName (addr + 1) values
        in

        let aw = Lp.bits (List.length vs) in
        write ("initial " ^ name ^ " = " ^ zeros dw ^ ";\n");
        write ("always @ (posedge " ^ cd ^ "_c)\n");
        write ("  if (" ^ input 0 ^ " == 1'b1)\n");
        write ("    case (" ^ input 1 ^ ")\n");
        rom1   "      " aw dw name 0 vs;
        write ("      default : " ^ name ^ " <= 0;\n");
        write ("    endcase\n");

    | Lp.VecRamMem (dw, vs) ->
        write ("initial begin\n");
        List2.iteri (fun index value -> write ("  " ^ name ^ "[" ^ string_of_int index ^ "] = " ^ string_of_int dw ^ "'b" ^ Intbig.binary_string_of_intbig value (Intbig.intbig_of_int dw) ^ ";\n")) vs;
        write ("end\n");

    | Lp.VecRamWrite (cd, _, _) ->
        write ("always @ (posedge " ^ cd ^ "_c)\n");
        write ("  if (" ^ input 1 ^ " == 1'b1)\n");
        write ("    " ^ input 0 ^ "[" ^ input 2 ^ "] <= " ^ input 3 ^ ";\n");

    | Lp.VecRamRead (cd, rbw, aw, dw) ->
        if rbw then begin
          write ("initial\n");
          write ("  " ^ name ^ " = " ^ zeros dw ^ ";\n");
          write ("always @ (posedge " ^ cd ^ "_c)\n");
          write ("  if (" ^ input 1 ^ " == 1'b1)\n");
          write ("    " ^ name ^ " <= " ^ input 0 ^ "[" ^ input 2 ^ "];\n");
        end else begin
          write ("initial\n");
          write ("  " ^ name ^ "a = " ^ zeros aw ^ ";\n");
          write ("always @ (posedge " ^ cd ^ "_c)\n");
          write ("  if (" ^ input 1 ^ " == 1'b1)\n");
          write ("    " ^ name ^ "a <= " ^ input 2 ^ ";\n");
          write ("assign " ^ name ^ " = " ^ input 0 ^ "[" ^ name ^ "a];\n");
        end
;;


