(*
    Confluence System Design Language Compiler
    Copyright (C) 2003-2004 Tom Hawkins (tomahawkins@yahoo.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*)


(* Misc helpers. *)

let zeros width =
  let rec f width =
    if width <= 0 then "" else "0" ^ f (width - 1)
  in
    "\"" ^ f width ^ "\"";;
  
let zzzs width =
  let rec f width =
    if width <= 0 then "" else "Z" ^ f (width - 1)
  in
    "\"" ^ f width ^ "\"";;

let write = Lp.write;;

let unsignedDecl width =
  "unsigned(" ^ string_of_int (width - 1) ^ " downto 0)";;

let rec codeParams params =
  match params with
    [] -> ""
  | param :: params -> ", " ^ string_of_int param ^ codeParams params;;
  

(* Declaration coder. *)

let decl lp name =
  if Lp.isCoded lp then begin
    match (lp, Lp.width lp) with
    | (Lp.VecInput _, _) | (Lp.VecOutput _, _) | (Lp.VecInOut _, _) | (Lp.VecAssert _, _) | (Lp.VecPrint _, _) -> ()

    | (Lp.VecReg (_, w), _) ->
        write ("signal " ^ name ^ " : " ^ unsignedDecl w ^ " := " ^ zeros w ^ ";\n");

    | (Lp.VecRom _, w) ->
        write ("signal " ^ name ^ " : " ^ unsignedDecl w ^ " := " ^ zeros w ^ ";\n");

    | (Lp.VecRamMem (dw, vs), _) ->
        let total = List.length vs in
        write ("type   " ^ name ^ "t is array (" ^ string_of_int (List.length vs - 1) ^ " downto 0) of " ^ unsignedDecl dw ^ ";\n");
        write ("signal " ^ name ^ " : " ^ name ^ "t := (\n");
        List2.iteri (fun index value -> write ("  " ^ string_of_int index ^ " => \"" ^ Intbig.binary_string_of_intbig value (Intbig.intbig_of_int dw) ^ "\"" ^ (if index = total - 1 then ");\n" else ",\n"))) vs

    | (Lp.VecRamRead (_, rbw, aw, dw), _) ->
        if rbw then begin
          write ("signal " ^ name ^ " : " ^ unsignedDecl dw ^ " := " ^ zeros dw ^ ";\n");
        end else begin
          write ("signal " ^ name ^ " : " ^ unsignedDecl dw ^ ";\n");
          write ("signal " ^ name ^ "a : " ^ unsignedDecl aw ^ " := " ^ zeros aw ^ ";\n");
        end

    | (Lp.VecRamWrite _, _) -> ()

    | (_, w) ->
        write ("signal " ^ name ^ " : " ^ unsignedDecl w ^ ";\n");
  end;;


(* Function coder. *)
                    
let func lp name inputs =

  if Lp.arity lp <> List.length inputs then raise (Lp.Error ("LpVerilog.func: Arity does not match input list length.  Primitive: " ^ Lp.primString lp ^ "  Inputs: " ^ string_of_int (List.length inputs)));

  let input n =
    List.nth inputs n
  in

  if Lp.isCoded lp then
    match lp with 
    | Lp.VecInput _ -> ()
  
    | Lp.VecOutput (n, _, w) ->
        write (n ^ "_o <= " ^ input 0 ^ ";\n")
  
    | Lp.VecInOut (n, _, w) ->
        write (n ^ "_b <= " ^ input 1 ^ " when " ^ input 0 ^ " = \"1\" else " ^ zzzs w ^ ";\n")
  
    | Lp.VecAssert msg ->
        write ("process (" ^ input 0 ^ ") begin\n");
        write ("  assert " ^ input 0 ^ " = \"1\"\n");
        write ("    report \"**  Assertion Failure\\n\\n    " ^ Lp.formatMsg msg ^ "\\n\\n\"\n");
        write ("    severity error;\n");
        write ("end process;\n");

    | Lp.VecPrint w ->
        let rec f name bh bl =
          if bl <= 0 then
            write ("character'val(to_integer(" ^ name ^ "(" ^ string_of_int bh ^ " downto " ^ string_of_int bl ^ ")))")
          else begin
            write ("character'val(to_integer(" ^ name ^ "(" ^ string_of_int bh ^ " downto " ^ string_of_int bl ^ "))), ");
            f name (bl - 1) (bl - 8);
          end
        in
        write ("process (" ^ input 0 ^ ", " ^ input 1 ^ ") begin\n");
        write ("  if " ^ input 0 ^ " = \"1\" then\n");
        write ("    report (");
        f (input 1) (w - 1) ((Lp.bytes w - 1) * 8);
        write (");\n");
        write ("  end if;\n");
        write ("end process;\n");
  
    | Lp.VecConst s ->
        write (name ^ " <= \"" ^ s ^ "\";\n")
        
    | Lp.VecBuf _ ->
        write (name ^ " <= " ^ input 0 ^ ";\n")
  
    | Lp.VecNot _ ->
        write (name ^ " <= not " ^ input 0 ^ ";\n")
  
    | Lp.VecAnd _ ->
        write (name ^ " <= " ^ input 0 ^ " and " ^ input 1 ^ ";\n")
  
    | Lp.VecOr _ ->
        write (name ^ " <= " ^ input 0 ^ " or " ^ input 1 ^ ";\n")
  
    | Lp.VecXor _ ->
        write (name ^ " <= " ^ input 0 ^ " xor " ^ input 1 ^ ";\n")
  
    | Lp.VecConcat (0, 0) -> ()
    | Lp.VecConcat (_, 0) ->
        write (name ^ " <= " ^ input 0 ^ ";\n")
    | Lp.VecConcat (0, _) ->
        write (name ^ " <= " ^ input 1 ^ ";\n")
    | Lp.VecConcat (_, _) ->
        write (name ^ " <= " ^ input 0 ^ " & "  ^ input 1 ^ ";\n")
  
    | Lp.VecSelect (_, bs) ->
        let rec codeSelect name bits =
          match bits with
            [] -> ""
          | bit :: bits ->
              " &\n  " ^ name ^ "(" ^ string_of_int bit ^ " downto " ^ string_of_int bit ^ ")" ^ codeSelect name bits
        in

        let b = string_of_int (List.hd bs) in
        let t = List.tl bs in

        if List.length bs = 1 then
          write (name ^ " <= " ^ input 0 ^ "(" ^ b ^ " downto " ^ b ^ ");\n")
        else
          write (name ^ " <= " ^ input 0 ^ "(" ^ b ^ " downto " ^ b ^ ")" ^ codeSelect (input 0) t ^ ";\n")
  
    | Lp.VecEqu _ ->
        write (name ^ " <= \"1\" when " ^ input 0 ^ " = " ^ input 1 ^ " else \"0\";\n")
 
    | Lp.VecAdd _ ->
        write (name ^ " <= " ^ input 0 ^ " + " ^ input 1 ^ ";\n")
  
    | Lp.VecSub _ ->
        write (name ^ " <= " ^ input 0 ^ " - " ^ input 1 ^ ";\n")
  
    | Lp.VecMul w ->
        write (name ^ " <= resize(" ^ input 0 ^ " * " ^ input 1 ^ ", " ^ string_of_int w ^ ");\n")
  
    | Lp.VecSwitch w ->
        write (name ^ " <= " ^ input 1 ^ " when " ^ input 0 ^ " = \"1\" else " ^ input 2 ^ ";\n");

    | Lp.VecExtComb (comp, params, wi, wo) ->
        write (name ^ "e : " ^ comp ^ " generic map (" ^ string_of_int wi ^ ", " ^ string_of_int wo ^ codeParams params ^ ") port map (" ^ input 0 ^ ", " ^ name ^ ");\n");

    | Lp.VecExtSequ (cd, comp, params, wi, wo) ->
        write (name ^ "e : " ^ comp ^ " generic map (" ^ string_of_int wi ^ ", " ^ string_of_int wo ^ codeParams params ^ ") port map (" ^ cd ^ "_c, " ^ input 0 ^ ", " ^ input 1 ^ ", " ^ input 2 ^ ", " ^ name ^ ");\n");

    | Lp.VecReg (cd, w) ->
        write ("process (" ^ cd ^ "_c) begin\n");
        write ("  if rising_edge(" ^ cd ^ "_c) then\n");
        write ("    if " ^ input 0 ^ " = \"1\" then\n");
        write ("      " ^ name ^ " <= " ^ zeros w ^ ";\n");
        write ("    elsif " ^ input 1 ^ " = \"1\" then\n");
        write ("      " ^ name ^ " <= " ^ input 2 ^ ";\n");
        write ("    end if;\n");
        write ("  end if;\n");
        write ("end process;\n");
  
    | Lp.VecRom (cd, dw, vs) ->
        let rec codeRomLines addrWidth dataWidth dataName addr values =
          match values with
            [] -> ()
          | value :: values ->
              write ("        when \"" ^ Lp.intToBin addr addrWidth ^ "\" => " ^ dataName ^ " <= \"" ^ Intbig.binary_string_of_intbig value (Intbig.intbig_of_int dw) ^ "\";\n");
              codeRomLines addrWidth dataWidth dataName (addr + 1) values
        in

        let aw = Lp.bits (List.length vs) in

        write ("process (" ^ cd ^ "_c) begin\n");
        write ("  if rising_edge(" ^ cd ^ "_c) then \n");
        write ("    if " ^ input 0 ^ " = \"1\" then\n");
        write ("      case " ^ input 1 ^ " is\n");
        codeRomLines aw dw name 0 vs;
        write ("        when others => " ^ name ^ " <= " ^ zeros dw ^ ";\n");
        write ("      end case;\n");
        write ("    end if;\n");
        write ("  end if;\n");
        write ("end process;\n");

    | Lp.VecRamMem _ -> ();
        
    | Lp.VecRamWrite (cd, _, _) ->
        write ("process (" ^ cd ^ "_c) begin\n");
        write ("  if rising_edge(" ^ cd ^ "_c) then\n");
        write ("    if " ^ input 1 ^ " = \"1\" then\n");
        write ("      " ^ input 0 ^ "(to_integer(" ^ input 2 ^ ")) <= " ^ input 3 ^ ";\n");
        write ("    end if;\n");
        write ("  end if;\n");
        write ("end process;\n");

    | Lp.VecRamRead (cd, rbw, _, _) ->
        if rbw then begin
          write ("process (" ^ cd ^ "_c) begin\n");
          write ("  if rising_edge(" ^ cd ^ "_c) then\n");
          write ("    if " ^ input 1 ^ " = \"1\" then\n");
          write ("      " ^ name ^ " <= " ^ input 0 ^ "(to_integer(" ^ input 2 ^ "));\n");
          write ("    end if;\n");
          write ("  end if;\n");
          write ("end process;\n");
        end else begin
          write ("process (" ^ cd ^ "_c) begin\n");
          write ("  if rising_edge(" ^ cd ^ "_c) then\n");
          write ("    if " ^ input 1 ^ " = \"1\" then\n");
          write ("      " ^ name ^ "a <= " ^ input 2 ^ ";\n");
          write ("    end if;\n");
          write ("  end if;\n");
          write ("end process;\n");
          write (name ^ " <= " ^ input 0 ^ "(to_integer(" ^ name ^ "a));\n");
        end;
;;

