(*
    Confluence System Design Language Compiler
    Copyright (C) 2003-2004 Tom Hawkins (tomahawkins@yahoo.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*)


(* Exception. *)

exception Error of string;;

(* Types. *)

type id = int;;

module OrderedId =
  struct
    type t = id
    let compare = compare
  end;;

module IdSet = Set.Make(OrderedId);;

module OrderedIdPort =
  struct
    type t = id * int
    let compare = compare
  end;;

module IdPortSet = Set.Make(OrderedIdPort);;

type space = {
          spaceIsRoot     : bool;
          spaceInstLoc    : string;
          spaceCompLoc    : string;
  mutable spaceParent     : space;
  mutable spaceSubs       : space list;
  mutable spacePrims      : IdSet.t;
  mutable spaceProducers  : IdSet.t;
  mutable spaceConsumers  : IdSet.t;
  mutable spaceInputs     : IdSet.t;
  mutable spaceOutputs    : IdSet.t;
};;

type nls = {
    mutable file        : string;
    mutable build       : string;
    mutable header      : string;
    c2p                 : (id * int, id) Hashtbl.t;
    p2cs                : (id, IdPortSet.t) Hashtbl.t;
    nodes               : (id, Lp.lp) Hashtbl.t;
    locs                : (id, Loc.loc list) Hashtbl.t;
    mutable inputs      : IdSet.t;   (* XXX Get rid of this. *)
    mutable outputs     : IdSet.t;   (* XXX Get rid of this. *)
    mutable inouts      : IdSet.t;   (* XXX Get rid of this. *)
    mutable asserts     : IdSet.t;   (* XXX Get rid of this. *)
    mutable memories    : IdSet.t;   (* XXX Get rid of this. *)
    mutable constants   : IdSet.t;   (* XXX Get rid of this. *)
    mutable nets        : IdSet.t list;
    clocks              : (id, string) Hashtbl.t;
    mutable nextId      : id;
    rootSpace           : space;
    mainSpace           : space;
    n2s                 : (id, space) Hashtbl.t;   (* Only used before propogating space information. *)
  };;


(** Space operations. *)

(** Return the name space. *)
let getSpaceMain nls =
  nls.mainSpace;;

let newSpaceRoot location =
  let rec space     = {
    spaceIsRoot     = true;
    spaceInstLoc    = location;
    spaceCompLoc    = location;
    spaceParent     = space;
    spaceSubs       = [];
    spacePrims      = IdSet.empty;
    spaceProducers  = IdSet.empty;
    spaceConsumers  = IdSet.empty;
    spaceInputs     = IdSet.empty;
    spaceOutputs    = IdSet.empty;
  } in
  space
;;

(** Create subspace. *)
let newSpaceSub parentSpace instLoc compLoc =
  {
    spaceIsRoot     = false;
    spaceInstLoc    = instLoc;
    spaceCompLoc    = compLoc;
    spaceParent     = parentSpace;
    spaceSubs       = [];
    spacePrims      = IdSet.empty;
    spaceProducers  = IdSet.empty;
    spaceConsumers  = IdSet.empty;
    spaceInputs     = IdSet.empty;
    spaceOutputs    = IdSet.empty;
  }
;;


(** Misc functions. *)

let fakeId = 0;;

let nextId nls =
  let id = nls.nextId
  in
    nls.nextId <- nls.nextId + 1;
    id;;

let idToInt id =
  id;;

let idPortToPair idPort =
  idPort;;

let info nls =
  "Nls Info\n" ^
  "file      : " ^ nls.file ^ "\n" ^
  "build     : " ^ nls.build ^ "\n" ^
  "header    : " ^ nls.header ^ "\n" ^
  "c2p       : " ^ Report.hashSize nls.c2p  ^ "\n" ^
  "p2cs      : " ^ Report.hashSize nls.p2cs ^ "\n" ^
  "nodes     : " ^ Report.hashSize nls.nodes ^ "\n" ^
  "inputs    : " ^ string_of_int (IdSet.cardinal nls.inputs) ^ "\n" ^
  "outputs   : " ^ string_of_int (IdSet.cardinal nls.outputs) ^ "\n" ^
  "inouts    : " ^ string_of_int (IdSet.cardinal nls.inouts) ^ "\n" ^
  "asserts   : " ^ string_of_int (IdSet.cardinal nls.asserts) ^ "\n" ^
  "memories  : " ^ string_of_int (IdSet.cardinal nls.memories) ^ "\n" ^
  "constants : " ^ string_of_int (IdSet.cardinal nls.constants) ^ "\n" ^
  "clocks    : " ^ Report.hashSize nls.clocks ^ "\n\n";;


(** Netlist operations. *)

let newNls () =
  let rootSpace = newSpaceRoot "root" in
  { 
    file       = "default_file";
    build      = "default_build";
    header     = "";
    c2p        = Hashtbl.create 2048;
    p2cs       = Hashtbl.create 2048;
    nodes      = Hashtbl.create 2048;
    locs       = Hashtbl.create 2048;
    inputs     = IdSet.empty;
    outputs    = IdSet.empty;
    inouts     = IdSet.empty;
    asserts    = IdSet.empty;
    memories   = IdSet.empty;
    constants  = IdSet.empty;
    nets       = [];
    clocks     = Hashtbl.create 1024;
    nextId     = 1;
    rootSpace  = rootSpace;
    mainSpace  = newSpaceSub rootSpace "main" "main";
    n2s        = Hashtbl.create 256;
  };;

let setFile   nls file   = nls.file   <- file;;
let setBuild  nls build  = nls.build  <- build;;
let setHeader nls header = nls.header <- header;;

let getFile   nls = nls.file;;
let getBuild  nls = nls.build;;
let getHeader nls = nls.header;;

let node nls id =
  try
    Hashtbl.find nls.nodes id
  with Not_found -> raise (Error "Node not found.");;

let nodeType nls id =
  Lp.primString (node nls id);;

let nodeArity nls id =
  Lp.arity (node nls id);;

let nodeWidth nls id =
  Lp.width (node nls id);;

let nodeInputWidths nls id =
  Lp.inputWidths (node nls id);;

let nodeLocs nls id =
  try
    Hashtbl.find nls.locs id
  with Not_found -> raise (Error "Node location not found.");;

let nodeList nls =
  let f id node idList =
    id :: idList
  in
    Hashtbl.fold f nls.nodes [];;

let nodeSet nls =
  let f id node idSet =
    IdSet.add id idSet
  in
    Hashtbl.fold f nls.nodes IdSet.empty;;

let contains nls id =
  Hashtbl.mem nls.nodes id;;

let source nls cId cPort =
  Hashtbl.find nls.c2p (cId, cPort);;

let sources nls cId =
  let arity = nodeArity nls cId
  in
    let rec build cPort =
      if cPort >= arity then IdSet.empty
      else
        IdSet.add (source nls cId cPort) (build (cPort + 1))
    in
      build 0;;

let sourceList nls cId =
  let arity = nodeArity nls cId
  in
    let rec build cPort =
      if cPort >= arity then []
      else
        source nls cId cPort :: (build (cPort + 1))
    in
      build 0;;

let sinks nls pId =
  Hashtbl.find nls.p2cs pId;;

let addConsumer nls cId cPort pId =
  Hashtbl.replace nls.p2cs pId (IdPortSet.add (cId, cPort) (sinks nls pId));;

let removeConsumer nls cId cPort =
  let pId = source nls cId cPort
  in
    Hashtbl.replace nls.p2cs pId (IdPortSet.remove (cId, cPort) (sinks nls pId));;

let reconnect nls cId cPort pId =
  removeConsumer nls cId cPort;
  addConsumer nls cId cPort pId;
  Hashtbl.replace nls.c2p (cId, cPort) pId;;

let redirect nls pIdOld pIdNew =
  let f (cId, cPort) =
    reconnect nls cId cPort pIdNew
  in
    IdPortSet.iter f (Hashtbl.find nls.p2cs pIdOld);;

let required nls cId =
  let rec f1 cPort arity sofar =
    if cPort < arity then
      f1 (cPort + 1) arity (IdSet.add (source nls cId cPort) sofar)
    else
      sofar
  in
  match node nls cId with
  | Lp.VecRamMem _ ->
      IdPortSet.fold (fun (id, _) writePorts ->
        match node nls id with
        | Lp.VecRamWrite _ -> IdSet.add id writePorts
        | _ -> writePorts
      ) (sinks nls cId) IdSet.empty
  | _ ->
      f1 0 (nodeArity nls cId) IdSet.empty
;; 

let rec delete nls id =
  let rec f1 cPort arity =
    if cPort < arity then begin
      removeConsumer nls id cPort;
      Hashtbl.remove nls.c2p (id, cPort);
      f1 (cPort + 1) arity
    end
  and f2 (cId, _) =
    delete nls cId
  in
    if contains nls id then begin
      f1 0 (nodeArity nls id);
      IdPortSet.iter f2 (sinks nls id);
      begin
        match node nls id with
        | Lp.VecInput   _  -> raise (Error "Illegal node deletion: Cannot delete input.")
        | Lp.VecOutput  _  -> raise (Error "Illegal node deletion: Cannot delete output.")
        | Lp.VecInOut   _  -> raise (Error "Illegal node deletion: Cannot delete inout.")
        | Lp.VecAssert  _  -> raise (Error "Illegal node deletion: Cannot delete assert.")
        | Lp.VecPrint   _  -> raise (Error "Illegal node deletion: Cannot delete print.")
        | Lp.VecReg     _  -> nls.memories  <- IdSet.remove id nls.memories
        | Lp.VecRom     _  -> nls.memories  <- IdSet.remove id nls.memories
        | Lp.VecRamRead _  -> nls.memories  <- IdSet.remove id nls.memories
        | Lp.VecConst   _  -> nls.constants <- IdSet.remove id nls.constants
        | _ -> ()
      end;
      Hashtbl.remove nls.nodes id;
      Hashtbl.remove nls.p2cs  id;
      Hashtbl.remove nls.n2s   id;
      Hashtbl.remove nls.locs  id;
    end
;;

let idCompare nls idA idB = (* General node order: inputs, outputs, inouts, any other. *)
  match (node nls idA, node nls idB) with
    (Lp.VecInput    (_, pA, _), Lp.VecInput    (_, pB, _)) -> compare pA pB
  | (Lp.VecInput    (_, pA, _), Lp.VecOutput   (_, pB, _)) -> compare pA pB
  | (Lp.VecInput    (_, pA, _), Lp.VecInOut    (_, pB, _)) -> compare pA pB
  | (Lp.VecOutput   (_, pA, _), Lp.VecInput    (_, pB, _)) -> compare pA pB
  | (Lp.VecOutput   (_, pA, _), Lp.VecOutput   (_, pB, _)) -> compare pA pB
  | (Lp.VecOutput   (_, pA, _), Lp.VecInOut    (_, pB, _)) -> compare pA pB
  | (Lp.VecInOut    (_, pA, _), Lp.VecInput    (_, pB, _)) -> compare pA pB
  | (Lp.VecInOut    (_, pA, _), Lp.VecOutput   (_, pB, _)) -> compare pA pB
  | (Lp.VecInOut    (_, pA, _), Lp.VecInOut    (_, pB, _)) -> compare pA pB
  | (_, _) -> (-1);;

let rec idInsert nls newId ids =
  match ids with
    [] -> newId :: []
  | id :: ids ->
      if idCompare nls newId id <= 0 then
        newId :: id :: ids
      else
        id :: idInsert nls newId ids;;

let portSorter nls id ids =
  match node nls id with
    Lp.VecInput    (_, _, w) -> if w > 0 then idInsert nls id ids else ids
  | Lp.VecOutput   (_, _, w) -> if w > 0 then idInsert nls id ids else ids
  | Lp.VecInOut    (_, _, w) -> if w > 0 then idInsert nls id ids else ids
  | _ -> raise (Error "Expecting Input, Output, or InOut in portList.");;

let portList nls = (* Returns a list of node ids, sorted. *)
  List.fold_left (fun ps p -> portSorter nls p ps) [] (IdSet.elements nls.inouts @ IdSet.elements nls.outputs @ IdSet.elements nls.inputs);;
    
let inputInOuts nls = (* Returns a list of node ids, sorted. *)
  List.fold_left (fun ps p -> portSorter nls p ps) [] (IdSet.elements nls.inouts @ IdSet.elements nls.inputs);;

let outputInOuts nls = (* Returns a list of node ids, sorted. *)
  List.fold_left (fun ps p -> portSorter nls p ps) [] (IdSet.elements nls.inouts @ IdSet.elements nls.outputs);;

let setClock nls id clock =
  Hashtbl.replace nls.clocks id clock;
  match node nls id with
  | Lp.VecReg (_, w)          -> Hashtbl.replace nls.nodes id (Lp.VecReg (clock, w))
  | Lp.VecRom (_, w, v)       -> Hashtbl.replace nls.nodes id (Lp.VecRom (clock, w, v))
  | Lp.VecRamWrite (_, a, d)      -> Hashtbl.replace nls.nodes id (Lp.VecRamWrite (clock, a, d))
  | Lp.VecRamRead  (_, rbw, a, d) -> Hashtbl.replace nls.nodes id (Lp.VecRamRead (clock, rbw, a, d))
  | Lp.VecExtSequ (_, name, params, din, dout) -> Hashtbl.replace nls.nodes id (Lp.VecExtSequ (clock, name, params, din, dout))
  | _ -> raise (Error "Expecting stateful.")
;;

let getClock nls id =
  try
    Hashtbl.find nls.clocks id
  with Not_found ->
    raise (Error "Clock domain not found.");;

let getClocks nls =
  let rec f1 ls i =
    match ls with
      [] -> [i]
    | l :: ls -> if l = i then l :: ls else l :: f1 ls i
  in
    let f2 id clock sofar =
      f1 sofar clock
    in
      Hashtbl.fold f2 nls.clocks [];;

let orderNodes nls =
  let rec build remaining computed order =
    if IdSet.is_empty remaining then
      List.rev order
    else
      let (ready, notReady) = IdSet.partition (function id -> IdSet.subset (sources nls id) computed) remaining in
      if IdSet.is_empty ready then raise (Error "orderNodes: Expecting ready nodes.  Received empty ready set.");
      build notReady (IdSet.union ready computed) (IdSet.elements ready @ order)
  in
    build (nodeSet nls) nls.memories []
;;


let orderStatefulNodes nls =
  let compareRamPorts a b =
    match (node nls a, node nls b) with
    | (Lp.VecRamRead (_, true,  _, _), _) -> (-1)
    | (Lp.VecRamRead (_, false, _, _), _) -> 1
    | (_, _) -> 0
  in
  let rec f1 nodes sofar =
    match nodes with
    | [] -> sofar
    | id :: nodes ->
        begin match node nls id with
          | Lp.VecReg _
          | Lp.VecRom _    -> f1 nodes (id :: sofar)
          | Lp.VecRamMem _ -> f1 nodes (List.sort compareRamPorts (List.map fst (IdPortSet.elements (sinks nls id))) @ sofar)
          | _              -> f1 nodes sofar
        end
  in
  f1 (nodeList nls) []
;;


let isValidTestBench nls =
  List.length (getClocks nls) = 1 &&
  IdSet.cardinal nls.outputs > 0 &&
  IdSet.cardinal nls.inputs = 0 &&
  IdSet.cardinal nls.inouts = 0;;

let isGeneratable nls =
  IdSet.cardinal nls.outputs + IdSet.cardinal nls.inouts + IdSet.cardinal nls.asserts > 0;;

let mergeNets nls id0 id1 =
  let rec f0 nets =
    match nets with
    | [] -> raise (Error "Net not found.")
    | net :: nets ->
        let has_0 = IdSet.mem id0 net in
        let has_1 = IdSet.mem id1 net in
        if has_0 && has_1 then  (* Nets already merged.*)
          net :: nets
        else if has_0 then f1 net id1 nets
        else if has_1 then f1 net id0 nets
        else net :: f0 nets
  and f1 found_net id nets =
    match nets with
    | [] -> raise (Error "Net not found.")
    | net :: nets ->
        if IdSet.mem id net then
          IdSet.union net found_net :: nets
        else
          net :: f1 found_net id nets
  in
    nls.nets <- f0 nls.nets
;;


(* Constructing new nodes: inputs, outputs, assertions, and primitives. *)

let newNode nls space locs node inputs =
  let id = nextId nls in
  let rec connect port inputs =
    match inputs with
      [] -> ()
    | input :: inputs ->
        Hashtbl.add nls.c2p (id, port) input;
        addConsumer nls id port input;
        connect (port + 1) inputs
  in
  Hashtbl.add nls.locs id locs;
  Hashtbl.add nls.p2cs id IdPortSet.empty;
  if not (Lp.isCoded node) then begin
    Hashtbl.add nls.nodes id (Lp.VecConst "");
    nls.constants <- IdSet.add id nls.constants;
  end else begin
    Hashtbl.add nls.n2s id space;
    Hashtbl.add nls.nodes id node;
    connect 0 inputs;
  end;
  id;;

let newVecInput nls space locs name position width =
  let id = newNode nls space locs (Lp.VecInput (name, position, width)) [] in
  nls.inputs <- IdSet.add id nls.inputs;
  id;;

let newVecOutput nls space locs name position width p0 =
  let id = newNode nls space locs (Lp.VecOutput (name, position, width)) [p0] in
  nls.outputs <- IdSet.add id nls.outputs;;

let newVecInOut nls space locs name position width p0 p1 =
  let id = newNode nls space locs (Lp.VecInOut (name, position, width)) [p0; p1] in
  nls.inouts <- IdSet.add id nls.inouts;
  id;;

let newVecAssert nls space locs test msg =
  let id = newNode nls space locs (Lp.VecAssert msg) [test] in
  nls.asserts <- IdSet.add id nls.asserts;;

let newVecPrint nls space locs enable msg =
  let id = newNode nls space locs (Lp.VecPrint (nodeWidth nls msg)) [enable; msg] in
  nls.asserts <- IdSet.add id nls.asserts;;

let newVecConst nls space locs value = 
  let id = newNode nls space locs (Lp.VecConst value) [] in
  if String.length value > 0 then nls.constants <- IdSet.add id nls.constants;
  id;;

let newVecBuf nls space locs p0 =
  newNode nls space locs (Lp.VecBuf (nodeWidth nls p0)) [p0];;

let newVecNot nls space locs p0 =
  newNode nls space locs (Lp.VecNot (nodeWidth nls p0)) [p0];;

let newVecAnd nls space locs p0 p1 =
  newNode nls space locs (Lp.VecAnd (nodeWidth nls p0)) [p0; p1];;

let newVecOr nls space locs p0 p1 =
  newNode nls space locs (Lp.VecOr (nodeWidth nls p0)) [p0; p1];;

let newVecXor nls space locs p0 p1 =
  newNode nls space locs (Lp.VecXor (nodeWidth nls p0)) [p0; p1];;

let newVecConcat nls space locs p0 p1 =
  newNode nls space locs (Lp.VecConcat (nodeWidth nls p0, nodeWidth nls p1)) [p0; p1];;

let newVecSelect nls space locs p0 bits =
  let rec checkBits width bits =
    match bits with
      [] -> ()
    | bit :: bits ->
        if bit < 0      then Report.error "Illegal bit selection: Bit select cannot be < 0.";
        if bit >= width then Report.error "Illegal bit selection: Bit select cannot exceed width - 1.";
        checkBits width bits
  in
    let inWidth = nodeWidth nls p0
    in
      checkBits inWidth bits;
      newNode nls space locs (Lp.VecSelect (inWidth, bits)) [p0];;

let newVecEqu nls space locs p0 p1 =
  newNode nls space locs (Lp.VecEqu (nodeWidth nls p0)) [p0; p1];;

let newVecAdd nls space locs p0 p1 =
  newNode nls space locs (Lp.VecAdd (nodeWidth nls p0)) [p0; p1];;

let newVecSub nls space locs p0 p1 =
  newNode nls space locs (Lp.VecSub (nodeWidth nls p0)) [p0; p1];;

let newVecMul nls space locs p0 p1 =
  newNode nls space locs (Lp.VecMul (nodeWidth nls p0)) [p0; p1];;

let newVecSwitch nls space locs pred t f =
  newNode nls space locs (Lp.VecSwitch (nodeWidth nls t)) [pred; t; f];;

let newVecExtComb nls space locs name params dataIn widthOut =
  newNode nls space locs (Lp.VecExtComb (name, params, (nodeWidth nls dataIn), widthOut)) [dataIn];;

let newVecExtSequ nls space locs name params reset enable dataIn widthOut =
  let id = newNode nls space locs (Lp.VecExtSequ ("clock", name, params, (nodeWidth nls dataIn), widthOut)) [reset; enable; dataIn] in
  nls.memories <- IdSet.add id nls.memories;
  id;;

let newVecReg nls space locs reset enable data =
  let id = newNode nls space locs (Lp.VecReg ("clock", nodeWidth nls data)) [reset; enable; data] in
  if nodeWidth nls data > 0 then nls.memories <- IdSet.add id nls.memories;
  id;;

let newVecRom nls space locs enable width values addr =
  let id = newNode nls space locs (Lp.VecRom ("clock", width, values)) [enable; addr] in
  if width > 0 then nls.memories <- IdSet.add id nls.memories;
  id;;

let newVecRamMem nls space locs width values =
  newNode nls space locs (Lp.VecRamMem (width, values)) [];;

let newVecRamWrite nls space locs mem enable addr data =
  let id = newNode nls space locs (Lp.VecRamWrite ("clock", nodeWidth nls addr, nodeWidth nls data)) [mem; enable; addr; data] in
  id
;;

let newVecRamRead nls space locs rbw mem enable addr =
  let (aw, dw) = match node nls mem with Lp.VecRamMem (dw, vs) -> (Lp.bits (List.length vs), dw) | _ -> raise (Error "Expecting memory link.") in
  let id = newNode nls space locs (Lp.VecRamRead ("clock", rbw, aw, dw)) [mem; enable; addr] in
  if dw > 0 then nls.memories <- IdSet.add id nls.memories;
  id
;; 

