(*
    Confluence System Design Language Compiler
    Copyright (C) 2003-2004 Tom Hawkins (tomahawkins@yahoo.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*)


(* Exception. *)

exception Error of string;;

(* Types. *)

module IdSet : Set.S;;
type id = IdSet.elt;;

(*
module IdPortSet : Set.S with type elt = id * int;;
*)
module IdPortSet : Set.S;;

type space = {
          spaceIsRoot     : bool;
          spaceInstLoc    : string;
          spaceCompLoc    : string;
  mutable spaceParent     : space;
  mutable spaceSubs       : space list;
  mutable spacePrims      : IdSet.t;
  mutable spaceProducers  : IdSet.t;
  mutable spaceConsumers  : IdSet.t;
  mutable spaceInputs     : IdSet.t;
  mutable spaceOutputs    : IdSet.t;
};;

type nls = {
    mutable file        : string;
    mutable build       : string;
    mutable header      : string;
    c2p                 : (id * int, id) Hashtbl.t;
    p2cs                : (id, IdPortSet.t) Hashtbl.t;
    nodes               : (id, Lp.lp) Hashtbl.t;
    locs                : (id, Loc.loc list) Hashtbl.t;
    mutable inputs      : IdSet.t;
    mutable outputs     : IdSet.t;
    mutable inouts      : IdSet.t;
    mutable asserts     : IdSet.t;
    mutable memories    : IdSet.t;
    mutable constants   : IdSet.t;
    mutable nets        : IdSet.t list;
    clocks              : (id, string) Hashtbl.t;
    mutable nextId      : id;
    rootSpace           : space;
    mainSpace           : space;
    n2s                 : (id, space) Hashtbl.t;
  };;


(* Misc. *)

val fakeId       : id;;
val idToInt      : id -> int;;
val idPortToPair : IdPortSet.elt -> id * int;;
val info         : nls -> string;;


(* Space functions. *)

val getSpaceMain : nls -> space;;
val newSpaceSub  : space -> string -> string -> space;;


(* Functions for setting and retrieving info and graph manipulation. *)

val newNls             : unit -> nls;;
val setFile            : nls -> string -> unit;;
val setBuild           : nls -> string -> unit;;
val setHeader          : nls -> string -> unit;;
val getFile            : nls -> string;;
val getBuild           : nls -> string;;
val getHeader          : nls -> string;;
val node               : nls -> id -> Lp.lp;;
val nodeType           : nls -> id -> string;;
val nodeArity          : nls -> id -> int;;
val nodeWidth          : nls -> id -> int;;
val nodeInputWidths    : nls -> id -> int list;;
val nodeLocs           : nls -> id -> Loc.loc list;;
val nodeList           : nls -> id list;;
val nodeSet            : nls -> IdSet.t;;
val contains           : nls -> id -> bool;;
val source             : nls -> id -> int -> id;;
val sources            : nls -> id -> IdSet.t;;
val sourceList         : nls -> id -> id list;;
val sinks              : nls -> id -> IdPortSet.t;;
val reconnect          : nls -> id -> int -> id -> unit;;
val redirect           : nls -> id -> id -> unit;;
val required           : nls -> id -> IdSet.t;;
val delete             : nls -> id -> unit;;
val portList           : nls -> id list;;
val inputInOuts        : nls -> id list;;
val outputInOuts       : nls -> id list;;
val setClock           : nls -> id -> string -> unit;;
val getClock           : nls -> id -> string;;
val getClocks          : nls -> string list;;
val orderNodes         : nls -> id list;;
val orderStatefulNodes : nls -> id list;;
val isValidTestBench   : nls -> bool;;
val isGeneratable      : nls -> bool;;
val mergeNets          : nls -> id -> id -> unit;;


(* Constructing new nodes including inputs, outputs, and asserts. *)

val newVecInput    : nls -> space -> Loc.loc list -> string -> int -> int -> id;;
val newVecOutput   : nls -> space -> Loc.loc list -> string -> int -> int -> id -> unit;;
val newVecInOut    : nls -> space -> Loc.loc list -> string -> int -> int -> id -> id -> id;;
val newVecAssert   : nls -> space -> Loc.loc list -> id -> string -> unit;;
val newVecPrint    : nls -> space -> Loc.loc list -> id -> id -> unit;;
val newVecConst    : nls -> space -> Loc.loc list -> string -> id;;
val newVecBuf      : nls -> space -> Loc.loc list -> id -> id;;
val newVecNot      : nls -> space -> Loc.loc list -> id -> id;;
val newVecAnd      : nls -> space -> Loc.loc list -> id -> id -> id;;
val newVecOr       : nls -> space -> Loc.loc list -> id -> id -> id;;
val newVecXor      : nls -> space -> Loc.loc list -> id -> id -> id;;
val newVecConcat   : nls -> space -> Loc.loc list -> id -> id -> id;;
val newVecSelect   : nls -> space -> Loc.loc list -> id -> int list -> id;;
val newVecEqu      : nls -> space -> Loc.loc list -> id -> id -> id;;
val newVecAdd      : nls -> space -> Loc.loc list -> id -> id -> id;;
val newVecSub      : nls -> space -> Loc.loc list -> id -> id -> id;;
val newVecMul      : nls -> space -> Loc.loc list -> id -> id -> id;;
val newVecSwitch   : nls -> space -> Loc.loc list -> id -> id -> id -> id;;
val newVecExtComb  : nls -> space -> Loc.loc list -> string -> int list -> id -> int -> id;;
val newVecExtSequ  : nls -> space -> Loc.loc list -> string -> int list -> id -> id -> id -> int -> id;;
val newVecReg      : nls -> space -> Loc.loc list -> id -> id -> id -> id;;
val newVecRom      : nls -> space -> Loc.loc list -> id -> int -> Intbig.intbig list -> id -> id;;
val newVecRamMem   : nls -> space -> Loc.loc list -> int -> Intbig.intbig list -> id;;
val newVecRamWrite : nls -> space -> Loc.loc list -> id -> id -> id -> id -> id;;
val newVecRamRead  : nls -> space -> Loc.loc list -> bool -> id -> id -> id -> id;;

