(*
    Confluence System Design Language Compiler
    Copyright (C) 2003-2004 Tom Hawkins (tomahawkins@yahoo.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*)


open Nls;;

let write = NlsCode.write;;
  

(* Misc helpers. *)

let bytes        = NlsCode.bytes;;
let nodeBytes    = NlsCode.nodeBytes;;
let byteRepeat   = NlsCode.byteRepeat;;
let copy         = NlsCode.copy;;
let nameByte     = NlsCode.nameByte;;
let intsToString = NlsCode.intsToString;;
let mask         = NlsCode.mask;;
let binToInt     = NlsCode.binToInt;;


(* Not helpers. *)

let rec codeNot1 name name0 byte =
  if byte < 0 then "" else
    "  " ^ nameByte name byte ^ " = 255 & ~" ^ nameByte name0 byte ^ ";\n" ^ codeNot1 name name0 (byte - 1);;

let codeNot name name0 width =
  let byte = bytes width - 1
  and mask = mask width
  in
    "  " ^ nameByte name byte ^ " = " ^ string_of_int mask ^ " & ~" ^ nameByte name0 byte ^ ";\n" ^ codeNot1 name name0 (byte - 1);;


(* Concat helpers. *)

let rec codeConcat3 name byte name0 byte0 w0 offset =
  if w0 <= 0 then "" else
    "  " ^ nameByte name byte ^ " = 255 & (" ^ nameByte name0 byte0 ^ " << " ^ string_of_int offset ^ " | " ^ nameByte name0 (byte0 - 1) ^ " >> " ^ string_of_int (8 - offset) ^ ");\n" ^
    codeConcat3 name (byte + 1) name0 (byte0 + 1) (w0 - 8) offset;;

let rec codeConcat2 name byte name0 byte0 w0 =
  if w0 <= 0 then "" else
    "  " ^ nameByte name byte ^ " = " ^ nameByte name0 byte0 ^ ";\n" ^ codeConcat2 name (byte + 1) name0 (byte0 + 1) (w0 - 8);;

let codeConcat1 name byte name0 w0 name1 byte1 w1 =
  if w1 = 0 then
    codeConcat2 name byte name0 0 w0
  else
    "  " ^ nameByte name byte ^ " = 255 & (" ^ nameByte name0 0 ^ " << " ^ string_of_int w1 ^ " | " ^ nameByte name1 byte1 ^ ");\n" ^
    codeConcat3 name (byte + 1) name0 1 (w0 - (8 - w1)) w1

let rec codeConcat name byte name0 w0 name1 byte1 w1 =
  if w1 < 8 then
    codeConcat1 name byte name0 w0 name1 byte1 w1
  else
    "  " ^ nameByte name byte ^ " = " ^ nameByte name1 byte1 ^ ";\n" ^ codeConcat name (byte + 1) name0 w0 name1 (byte1 + 1) (w1 - 8);;
  
(* Select helpers. *)

let codeSelect2 bit name0 selBit =
  let selByte = selBit / 8
  and selOffset = selBit mod 8
  in
    let shift =
      if bit = selOffset then "" else
        if bit > selOffset then " << " ^ string_of_int (bit - selOffset) else
          " >> " ^ string_of_int (selOffset - bit)
    in
      "(" ^ string_of_int (int_of_float (2.0 ** float_of_int bit)) ^ " & (" ^ nameByte name0 selByte ^ shift ^ "))";;

let rec codeSelect1 bit name0 selBits =
  match selBits with
    [] -> ";\n"
  | selBit :: []      -> codeSelect2 bit name0 selBit ^ ";\n"
  | selBit :: selBits -> codeSelect2 bit name0 selBit ^ " | " ^ codeSelect1 (bit + 1) name0 selBits;;

let rec codeSelect name byte name0 selBits =
  let l = List.length selBits
  in
    if l <= 0 then "" else
      if l < 8 then
        "  " ^ nameByte name byte ^ " = " ^ codeSelect1 0 name0 selBits
      else
        "  " ^ nameByte name byte ^ " = " ^ codeSelect1 0 name0 (List2.take selBits 8) ^ codeSelect name (byte + 1) name0 (List2.drop selBits 8);;

  
(* Equ helpers. *)

let rec codeEqu1 name0 name1 byte =
  let byte = byte - 1
  in
    "(" ^ nameByte name0 byte ^ " == " ^ nameByte name1 byte ^ ")"  ^ if byte <= 0 then "" else " & " ^ codeEqu1 name0 name1 byte;;

let codeEqu name name0 name1 bytes =
  "  " ^ nameByte name 0 ^ " = " ^ codeEqu1 name0 name1 bytes ^ ";\n";;


(* Add/Sub helpers. *)

(*
let rec codeAddSub1 name name0 name1 byte op mask firstByte =
  let lastByte = byte <= 0
  in
    "(tmp = (unsigned int) " ^ nameByte name0 byte ^ op ^ "(unsigned int) " ^ nameByte name1 byte ^ (if lastByte then "" else op ^ codeAddSub1 name name0 name1 (byte - 1) op "255" false) ^ ", " ^ nameByte name byte ^ " = " ^ mask ^ " & (unsigned char) tmp" ^ (if firstByte then "" else ", 1 & (tmp >> 8)") ^ ")";;
    *)

let codeAddSub name name0 name1 width op =
  let mask = string_of_int (mask width) in
  let b = bytes width in
  let op = " " ^ op ^ " " in
  let rec f b0 =
    let lastByte = b0 = (b - 1) in
      "    tmp = (unsigned short) " ^ nameByte name0 b0 ^ op ^ "(unsigned short) " ^ nameByte name1 b0 ^ (if b0 = 0 then "" else op ^ "(1 & tmp >> 8)") ^ ";\n" ^
      "    " ^ nameByte name b0 ^ " = " ^ (if lastByte then mask else "255") ^ " & (unsigned char) tmp;\n" ^ (if lastByte then "" else f (b0 + 1))
  in
    "  {\n    unsigned short tmp;\n" ^ f 0 ^ "  }\n";;


(* Mul helpers. *)

let codeMul name name0 name1 w =
  let b = bytes w in
  let mask = string_of_int (mask w) in

  let rec codeMul2 b0 b1 =
    let lastByte = b0 = 0 in
    "((unsigned short) " ^ nameByte name0 b0 ^ " * (unsigned short) " ^ nameByte name1 b1 ^ ")" ^ (if lastByte then "" else " + " ^ codeMul2 (b0 - 1) (b1 + 1))
  in

  let rec codeMul1 b0 =
    let firstByte = b0 = 0 in
    let lastByte = b0 = (b - 1) in
    "    tmp = " ^ codeMul2 b0 0 ^ (if firstByte then "" else " + (tmp >> 8)") ^ ";\n" ^
    "    " ^ nameByte name b0 ^ " = " ^ (if lastByte then mask else "255") ^ " & (unsigned char) tmp;\n" ^ (if lastByte then "" else codeMul1 (b0 + 1))
  in
  "  {\n    unsigned short tmp;\n" ^ codeMul1 0 ^ "  }\n"
;;


(* Reg helpers. *)

let codeRegCycle name width =
  let bytes = bytes width in
  "  {\n" ^
  "    char rn, e, en;\n" ^
  "    rn = ~ - " ^ name ^ "r[0];\n" ^
  "    e  = - " ^ name ^ "e[0];\n" ^
  "    en = ~ e;\n" ^
  byteRepeat (fun i -> "    " ^ name ^ i ^ " = (rn & e & " ^ name ^ "d" ^ i ^ ") + (rn & en & " ^ name ^ i ^ ");\n") bytes ^
  "  }\n";;


(* Ram/Rom helpers. *)

let rec codeRamRomIndex name byte =
  let lastByte = byte <= 0
  in
    "((unsigned int) " ^ nameByte name byte ^ (if lastByte then "" else " << " ^ string_of_int (8 * byte)) ^ ")" ^ (if lastByte then "" else " + " ^ codeRamRomIndex name (byte - 1));;

let rec codeRamRomOutput1 name memName byte =
  let lastByte = byte <= 0 in
  let nameWithByte = nameByte name byte in
  "    " ^ nameWithByte ^ " = (e & " ^ memName ^ "[base" ^ (if lastByte then "" else " + " ^ string_of_int byte) ^ "]) + (en & " ^ nameWithByte ^ ");\n" ^
  (if lastByte then "" else codeRamRomOutput1 name memName (byte - 1));;

let codeRamRomOutput name memName addrBytes dataBytes =
  "  {\n" ^
  "    char e, en;\n" ^
  "    int base;\n" ^
  "    e  = - " ^ name ^ "e[0];\n" ^
  "    en = ~ e;\n" ^
  "    base = " ^ string_of_int dataBytes ^ " * (" ^ codeRamRomIndex (name ^ "a") (addrBytes - 1) ^ ");\n" ^
  codeRamRomOutput1 name memName (dataBytes - 1) ^
  "  }\n";;

(* Rom helpers. *)

let codeRomCalc name addrBytes enable addrIn =
  copy "  " (name ^ "e") enable 1 ^
  copy "  " (name ^ "a") addrIn addrBytes;;

let codeRomCycle name addrBytes dataBytes =
  codeRamRomOutput name addrBytes dataBytes;;
  (*
  "  if (" ^ name ^ "e[0]) {\n" ^
  "    { int base = " ^ string_of_int dataBytes ^ " * (" ^ codeRamRomIndex (name ^ "a") (addrBytes - 1) ^ "); " ^ codeRamRomOutput name (dataBytes - 1) ^ " }\n" ^
  "  }\n"
  *)

(*
always @ (posedge clock_c)
  if (n2 == 1'b1)
    case (n13)
      2'b00 : n21 <= 10'b0100000000;
      2'b01 : n21 <= 10'b1000000001;
      2'b10 : n21 <= 10'b1100000010;
      2'b11 : n21 <= 10'b0000000011;
      default : n21 <= 0;
    endcase
    *)



  
(* Ram helpers. *)

let codeRamWriteCalc name aw dw enable addr data =
  copy "  " (name ^ "e") enable 1 ^
  copy "  " (name ^ "a") addr (bytes aw) ^
  copy "  " (name ^ "d") data (bytes dw)
;;

let codeRamReadCalc name aw enable addr =
  copy "  " (name ^ "e") enable 1 ^
  copy "  " (name ^ "a") addr (bytes aw)
;;

let rec codeRamCycle2 name memName byte =
  let lastByte = byte <= 0 in
  let cellName =  memName ^ "[base" ^ (if lastByte then "" else " + " ^ string_of_int byte) ^ "]" in
    "    " ^ cellName ^ " = (e & " ^ nameByte (name ^ "d") byte ^ ") + (en & " ^ cellName ^ ");\n" ^
    (if lastByte then "" else codeRamCycle2 name memName (byte - 1));;

let codeRamWriteCycle name memName addrBytes dataBytes =
  "  {\n" ^
  "    char e, en;\n" ^
  "    int base;\n" ^
  "    e  = - " ^ name ^ "e[0];\n" ^
  "    en = ~ e;\n" ^
  "    base = " ^ string_of_int dataBytes ^ " * (" ^ codeRamRomIndex (name ^ "a") (addrBytes - 1) ^ ");\n" ^
  codeRamCycle2 name memName (dataBytes - 1) ^
  "  }\n"
;;



(* Code nodes. *)

let codeNodeCalc nls id =
  if Lp.isCoded (node nls id) then
    let name = NlsCode.signalName nls id in
    let input n = NlsCode.producerSignalName nls id n in
    match node nls id with 
  
    | Lp.VecInput (_, _, w) ->
        if w mod 8 <> 0 then
          let msbyte = "[" ^ string_of_int (bytes w - 1) ^ "]" in
          write ("  " ^ name ^ msbyte ^ " = " ^ string_of_int (mask w) ^ " & " ^ name ^ msbyte ^ ";\n")
  
    | Lp.VecOutput (_, _, w) ->
        write (copy "  " name (input 0) (bytes w));
  
    | Lp.VecInOut (_, _, w) ->  (* XXX Mask off msbs.  Get rid of *. *)
        write ("  {\n");
        write ("    char s1, s0;\n");
        write ("    s1 = " ^ input 0 ^ "[0];\n");
        write ("    s0 = ! s1;\n");
        write (byteRepeat (fun i -> "    " ^ name ^ i ^ " = (s1 * " ^ input 1 ^ i ^ ") + (s0 * " ^ name ^ i ^ ");\n") (bytes w));
        write ("  }\n");
  
    | Lp.VecAssert msg ->
        if not (Lp.embeddedC ()) then begin
          write ("  if (!" ^ input 0 ^ "[0])\n");
          write ("    fprintf(stderr, \"**  Assertion Violation\\n\\n    " ^ Lp.formatMsg msg ^ "\\n\\n\");\n");
        end

    | Lp.VecPrint w ->
        let rec f name byte =
          if byte >= 0 then begin
            write (", " ^ name ^ "[" ^ string_of_int byte ^ "]");
            f name (byte - 1)
          end
        in
        if not (Lp.embeddedC ()) then begin
          write ("  if (" ^ input 0 ^ "[0])\n");
          write ("    fprintf(stderr, \"");
          for i = 1 to bytes w do write "%c" done;
          write ("\\n\"");
          f (input 1) (bytes w - 1);
          write (");\n");
        end
  
    | Lp.VecConst s -> () (* Handled in init. *)
        
    | Lp.VecBuf w ->
        write (copy "  " name (input 0) (bytes w));
  
    | Lp.VecNot w ->
        write (codeNot name (input 0) w)
  
    | Lp.VecAnd w ->
        let f byte = "  " ^ name ^ byte ^ " = " ^ input 0 ^ byte ^ " & " ^ input 1 ^ byte ^ ";\n" in
        write (byteRepeat f (bytes w));
  
    | Lp.VecOr w ->
        let f byte = "  " ^ name ^ byte ^ " = " ^ input 0 ^ byte ^ " | " ^ input 1 ^ byte ^ ";\n" in
        write (byteRepeat f (bytes w));
  
    | Lp.VecXor w ->
        let f byte = "  " ^ name ^ byte ^ " = " ^ input 0 ^ byte ^ " ^ " ^ input 1 ^ byte ^ ";\n" in
        write (byteRepeat f (bytes w));
  
    | Lp.VecConcat (0, 0) -> ()
    | Lp.VecConcat (w, 0) ->
        write (copy "  " name (input 0) (bytes w));
    | Lp.VecConcat (0, w) ->
        write (copy "  " name (input 1) (bytes w));
    | Lp.VecConcat (w0, w1) ->
        write (codeConcat name 0 (input 0) w0 (input 1) 0 w1);
  
    | Lp.VecSelect (_, bs) ->
        write (codeSelect name 0 (input 0) (List.rev bs));
  
    | Lp.VecEqu w ->
        write (codeEqu name (input 0) (input 1) (bytes w));
  
    | Lp.VecAdd w ->
        write (codeAddSub name (input 0) (input 1) w "+");
  
    | Lp.VecSub w ->
        write (codeAddSub name (input 0) (input 1) w "-");
  
    | Lp.VecMul w ->
        write (codeMul name (input 0) (input 1) w);
  
    | Lp.VecSwitch w ->
        write ("  {\n");
        write ("    char s1, s0;\n");
        write ("    s1 = - " ^ input 0 ^ "[0];\n");
        write ("    s0 = ~ s1;\n");
        write (byteRepeat (fun i -> "    " ^ name ^ i ^ " = (s1 & " ^ input 1 ^ i ^ ") + (s0 & " ^ input 2 ^ i ^ ");\n") (bytes w));
        write ("  }\n");

    | Lp.VecExtComb (extName, params, wi, wo) ->
        write ("  " ^ extName ^ "(" ^ string_of_int wi ^ ", " ^ string_of_int wo);
        List.iter (fun param -> write (", " ^ string_of_int param)) params;
        write (", " ^ input 0 ^ ", " ^ name ^ ");\n")

    | Lp.VecExtSequ (_, _, _, w, _) | Lp.VecReg (_, w) ->
        write (copy "  " (name ^ "r") (input 0) 1);
        write (copy "  " (name ^ "e") (input 1) 1);
        write (copy "  " (name ^ "d") (input 2) (bytes w))
  
    | Lp.VecRom (_, _, vs) ->
        let addrBytes = bytes (Lp.bits (List.length vs)) in
        write (codeRomCalc name addrBytes (input 0) (input 1));

    | Lp.VecRamMem _ -> ()

    | Lp.VecRamWrite (_, aw, dw) ->
        write (codeRamWriteCalc name aw dw (input 1) (input 2) (input 3))

    | Lp.VecRamRead (_, _, aw, _) ->
        write (codeRamReadCalc name aw (input 1) (input 2))
;;




let codeNodeCycle nls id =
  if Lp.isCoded (node nls id) then
    let name = NlsCode.signalName nls id in
    let input n = NlsCode.producerSignalName nls id n in
    match node nls id with 

      Lp.VecExtSequ (_, extName, params, wi, wo) ->
        write ("  " ^ extName ^ "(" ^ string_of_int wi ^ ", " ^ string_of_int wo);
        List.iter (fun param -> write (", " ^ string_of_int param)) params;
        write (", " ^ name ^ "r[0], " ^ name ^ "e[0], " ^ name ^ "d, " ^ name ^ ");\n")

    | Lp.VecReg (_, w) ->
        write (codeRegCycle name w);
  
    | Lp.VecRom (_, _, vs) ->
        write (codeRomCycle name (name ^ "m") (bytes (Lp.bits (List.length vs))) (nodeBytes nls id));

    | Lp.VecRamWrite (_, aw, dw) ->
        write (codeRamWriteCycle name (input 0) (bytes aw) (bytes dw))
  
    | Lp.VecRamRead (_, _, aw, dw) ->
        write (codeRamRomOutput name (input 0) (bytes aw) (bytes dw))
  
    | _ -> raise (Error "Invalid node for cycle.  Requires memory.")
;;

  


(* Declaration coding. *)

let codeNodeDeclName name bytes =
  write ("static unsigned char " ^ name ^ "[" ^ string_of_int bytes ^ "];\n");;

let codeNodeDecl nls id =
  if Lp.isCoded (node nls id) then
    let name = NlsCode.signalName nls id in
    let bs = nodeBytes nls id in
    match node nls id with
      Lp.VecInput _  -> write ("static unsigned char *" ^ name ^ ";\n");
    | Lp.VecOutput _ -> write ("static unsigned char *" ^ name ^ ";\n");
    | Lp.VecInOut _  -> write ("static unsigned char *" ^ name ^ ";\n");
    | Lp.VecAssert _ -> ()
    | Lp.VecPrint  _ -> ()

    | Lp.VecExtSequ (_, _, _, wi, wo) ->
        codeNodeDeclName name bs;
        codeNodeDeclName (name ^ "r") 1;
        codeNodeDeclName (name ^ "e") 1;
        codeNodeDeclName (name ^ "d") (bytes wi);

    | Lp.VecReg _ ->
        let dataBytes = bs in
        codeNodeDeclName name bs;
        codeNodeDeclName (name ^ "r") 1;
        codeNodeDeclName (name ^ "e") 1;
        codeNodeDeclName (name ^ "d") bs;

    | Lp.VecRom (_, _, vs) ->
        let memBytes = bs * List.length vs in
        let addrBytes = bytes (Lp.bits (List.length vs)) in
        let dataBytes = bs in
        codeNodeDeclName name bs;
        codeNodeDeclName (name ^ "e") 1;
        codeNodeDeclName (name ^ "a") addrBytes;
        codeNodeDeclName (name ^ "m") memBytes;

    | Lp.VecRamMem (dw, vs) ->
        let memBytes = bytes dw * List.length vs in
        codeNodeDeclName name memBytes;
        
    | Lp.VecRamWrite (_, aw, dw) ->
        let addrBytes = bytes aw in
        let dataBytes = bytes dw in
        codeNodeDeclName name dataBytes;
        codeNodeDeclName (name ^ "e") 1;
        codeNodeDeclName (name ^ "a") addrBytes;
        codeNodeDeclName (name ^ "d") dataBytes;

    | Lp.VecRamRead (_, _, aw, dw) ->
        let addrBytes = bytes aw in
        let dataBytes = bytes dw in
        codeNodeDeclName name dataBytes;
        codeNodeDeclName (name ^ "e") 1;
        codeNodeDeclName (name ^ "a") addrBytes;

    | _ -> 
        codeNodeDeclName name bs;
;;



(* Initialization coding. *)

let initZero name bytes =
  if bytes = 1 then
    "  " ^ name ^ "[0] = 0;\n"
  else
    "  { int index; for (index = " ^ string_of_int bytes ^ "; index > 0;) " ^ name ^ "[--index] = 0; }\n";;
    
let rec initMem1 name byte width value valueWidth values =
  if valueWidth <= 8 then
    "  " ^ name ^ "[" ^ string_of_int byte ^ "] = " ^ Intbig.string_of_intbig value ^ ";\n" ^
    initMem name (byte + 1) width values
  else
    "  " ^ name ^ "[" ^ string_of_int byte ^ "] = " ^
    Intbig.string_of_intbig (Intbig.bw_and (Intbig.intbig_of_string "255") value) ^ ";\n" ^
    initMem1 name (byte + 1) width (Intbig.shift_right value (Intbig.intbig_of_string "8")) (valueWidth - 8) values

and initMem name byte width values =      (* XXX May be benificial to bring in write for large roms. *)
  match values with
    [] -> ""
  | value :: values ->
      initMem1 name byte width value width values;;

let codeNodeInit nls id =
  if Lp.isCoded (node nls id) then
    let name = NlsCode.signalName nls id in
    let bs = nodeBytes nls id in
    match node nls id with
      Lp.VecInput _  -> ()
    | Lp.VecOutput _ -> ()
    | Lp.VecInOut _  -> ()
    | Lp.VecAssert _ -> ()
    | Lp.VecPrint  _ -> ()

    | Lp.VecConst s ->
        let rec codeConst byte bits =
          let w = String.length bits in
          if w = 0 then () else
            if w < 8 then
              write ("  " ^ nameByte name byte ^ " = " ^ string_of_int (binToInt bits) ^ ";\n")
            else begin
              write ("  " ^ nameByte name byte ^ " = " ^ string_of_int (binToInt (String.sub bits (w - 8) 8)) ^ ";\n");
              codeConst (byte + 1) (String.sub bits 0 (w - 8))
            end
        in
        codeConst 0 s

    | Lp.VecExtSequ (_, _, _, wi, wo) ->
        write (initZero name bs);
        write (initZero (name ^ "r") 1);
        write (initZero (name ^ "e") 1);
        write (initZero (name ^ "d") (bytes wi));
    
    | Lp.VecReg _ ->
        let dataBytes = bs in
        write (initZero name dataBytes);
        write (initZero (name ^ "r") 1);
        write (initZero (name ^ "e") 1);
        write (initZero (name ^ "d") dataBytes);

    | Lp.VecRom (_, w, vs) ->
        let memBytes = bs * List.length vs in
        let addrBytes = bytes (Lp.bits (List.length vs)) in
        let dataBytes = bs in
        write (initZero name dataBytes);
        write (initZero (name ^ "e") 1);
        write (initZero (name ^ "a") addrBytes);
        write (initMem (name ^ "m") 0 w vs);

    | Lp.VecRamMem (dw, vs) ->
        let memBytes = bytes dw * List.length vs in
        let addrBytes = bytes (Lp.bits (List.length vs)) in
        let dataBytes = bytes dw in
        write (initMem name 0 dw vs);

    | Lp.VecRamWrite (_, aw, dw) ->
        let addrBytes = bytes aw in
        let dataBytes = bytes dw in
        write (initZero name dataBytes);
        write (initZero (name ^ "e") 1);
        write (initZero (name ^ "a") addrBytes);
        write (initZero (name ^ "d") dataBytes);

    | Lp.VecRamRead (_, _, aw, dw) ->
        let addrBytes = bytes aw in
        let dataBytes = bytes dw in
        write (initZero name dataBytes);
        write (initZero (name ^ "e") 1);
        write (initZero (name ^ "a") addrBytes);

    | _ -> 
        write (initZero name bs);
;;









let codeArgs nls =
  let rec f ids =
    match ids with
      [] -> ""
    | id :: [] -> "unsigned char* port_" ^ NlsCode.signalName nls id
    | id :: ids -> "unsigned char* port_" ^ NlsCode.signalName nls id ^ ", " ^ f ids
  in
    "(" ^ f (portList nls) ^ ")";;

(* Header *)

let codeHeader nls buildName =
  "#ifdef __cplusplus\nextern \"C\" {\n#endif\n\n" ^
  "void " ^ buildName ^ "_ports" ^ codeArgs nls ^ ";\n" ^
  "void " ^ buildName ^ "_init();\n" ^
  "void " ^ buildName ^ "_calc();\n" ^
  List.fold_left (fun sofar clock -> sofar ^ "void " ^ buildName ^ "_cycle_" ^ clock ^ "();\n") "" (getClocks nls) ^
  (if not (Lp.embeddedC ()) then
    "void " ^ buildName ^ "_sim_init(const char* file);\n" ^
    "void " ^ buildName ^ "_sim_end();\n" ^
    "void " ^ buildName ^ "_sim_sample();\n\n"
   else
    "\n") ^
  "#ifdef __cplusplus\n}\n#endif\n\n";;


(* Main C functions. *)

let codeDecl nls buildName nodes =
  List.iter (fun id -> codeNodeDecl nls id) nodes;
  write "\n";;

let codeInit nls buildName nodes =
  write ("void " ^ buildName ^ "_init() {\n");
  List.iter (fun id -> codeNodeInit nls id) nodes;
  write ("  " ^ buildName ^ "_calc();\n");
  write "}\n\n";;

let codePorts nls buildName =
  write ("void " ^ buildName ^ "_ports" ^ codeArgs nls ^ " {\n");
  List.iter (fun id -> let name = NlsCode.signalName nls id in write ("  " ^ name ^ " = port_" ^ name ^ ";\n")) (portList nls);
  write "}\n\n";;

let codeCalc nls buildName nodes =
  write ("void " ^ buildName ^ "_calc() {\n");
  List.iter (fun id -> codeNodeCalc nls id) nodes;
  write "}\n\n";;

let codeCycles nls buildName =
  let f1 clock =
    write ("void " ^ buildName ^ "_cycle_" ^ clock ^ "()\n{\n");
    List.iter (fun id -> if Lp.clockDomain (node nls id) = clock then codeNodeCycle nls id) (orderStatefulNodes nls);
    write ("}\n\n");
  in
    List.iter f1 (getClocks nls);;


(* VCD functions. *)

let codeSimLocals nls =
  write ("static FILE* sim_file;\n");
  write ("static int sim_count;\n");
  List.iter (fun id -> write ("static unsigned char sim_" ^ NlsCode.signalName nls id ^ "[" ^ string_of_int (nodeBytes nls id) ^ "];\n")) (portList nls);
  write ("\n");;

let rec intToVcdId i =
  let c = Char.chr (i mod 94 + 33)
  in
    let s =
      match c with
        '\\' -> "\\\\"
      | '"'  -> "\\\""
      | '%'  -> "%%"
      | _ -> String.make 1 c
    in
      s ^ (if i / 94 = 0 then "" else intToVcdId (i / 94));;

let codeSimPrintVcd indent name width bytes num =
  let vcdid = intToVcdId num
  in
    if width = 1 then begin
      write (indent ^ "fprintf(sim_file, " ^ name ^ "[0] ? \"1" ^ vcdid ^ "\\n\" : \"0" ^ vcdid ^ "\\n\");\n");
    end else begin
      write (indent ^ "{\n");
      write (indent ^ "  int bit;\n");
      write (indent ^ "  int byte;\n");
      write (indent ^ "  int found = 0;\n");
      write (indent ^ "  for (byte = " ^ string_of_int (bytes - 1) ^ "; byte >= 0; byte--) {\n");
      write (indent ^ "    for (bit = 128; bit >= 1; bit /= 2) {\n");
      write (indent ^ "      if (found)\n");
      write (indent ^ "        fprintf(sim_file, (bit & " ^ name ^ "[byte]) ? \"1\" : \"0\");\n");
      write (indent ^ "      else if (bit & " ^ name ^ "[byte]) {\n");
      write (indent ^ "        found = 1;\n");
      write (indent ^ "        fprintf(sim_file, \"b1\");\n");
      write (indent ^ "      }\n");
      write (indent ^ "    }\n");
      write (indent ^ "  }\n");
      write (indent ^ "  if (!found)\n");
      write (indent ^ "    fprintf(sim_file, \"b0\");\n");
      write (indent ^ "  fprintf(sim_file, \" " ^ vcdid ^ "\\n\");\n");
      write (indent ^ "}\n");
    end;;

let rec codeSimInit1 nls ids num =
  match ids with
    [] -> ()
  | id :: ids ->
      write ("    \"$var wire " ^ string_of_int (nodeWidth nls id) ^ " " ^ intToVcdId num ^ " " ^ NlsCode.signalName nls id ^ " $end\\n\"\n");
      codeSimInit1 nls ids (num + 1);;

let rec codeSimInitSample nls ids num =
  match ids with
    [] -> ()
  | id :: ids ->
      codeSimPrintVcd "  " (NlsCode.signalName nls id) (nodeWidth nls id) (nodeBytes nls id) num;
      codeSimInitSample nls ids (num + 1);;

let codeSimInit nls buildName =
  let ports = portList nls
  in
    write ("void " ^ buildName ^ "_sim_init(const char* file) {\n");
    write ("  time_t date = time(NULL);\n");
    write ("  sim_count = 1;\n");
    write ("  " ^ buildName ^ "_init();\n");
    List.iter (fun id -> write (copy "  " ("sim_" ^ NlsCode.signalName nls id) (NlsCode.signalName nls id) (nodeBytes nls id))) ports; 
    write ("  sim_file = fopen(file, \"w\");\n");
    write ("  fprintf(sim_file,\n");
    write ("    \"$date\\n  %s$end\\n\"\n");
    write ("    \"$version\\n  Confluence System Design Lanugage " ^ Version.version ^ " -- http://www.launchbird.com/\\n$end\\n\"\n");
    write ("    \"$timescale\\n  1 ns\\n$end\\n\"\n");
    write ("    \"$scope module " ^ buildName ^ " $end\\n\"\n");
    codeSimInit1 nls ports 0;
    write ("    \"$upscope $end\\n\"\n");
    write ("    \"$enddefinitions $end\\n\"\n");
    write ("    \"#0\\n\"\n");
    write ("    \"$dumpvars\\n\", asctime(localtime(&date)));\n");
    codeSimInitSample nls ports 0;
    write ("  fprintf(sim_file, \"$end\\n\");\n");
    write ("}\n\n");;

let codeSimEnd buildName =
  write ("void " ^ buildName ^ "_sim_end() {\n");
  write ("  fprintf(sim_file, \"#%d\\n\", sim_count);\n");
  write ("  fclose(sim_file);\n");
  write ("}\n\n");;

let rec codeSimSample2 name byte =
  nameByte ("sim_" ^ name) byte ^ " != " ^ nameByte name byte ^ (if byte <= 0 then "" else " || " ^ codeSimSample2 name (byte - 1));;

let rec codeSimSample1 nls ids num =
  match ids with
    [] -> ()
  | id :: ids ->
      let bytes = nodeBytes nls id
      and name  = NlsCode.signalName nls id
      in
        write ("  if (" ^ codeSimSample2 name (bytes - 1) ^ ") {\n");
        write ("    if (!changed) { changed = 1; fprintf(sim_file, \"#%d\\n\", sim_count); }\n");
        codeSimPrintVcd "    " name (nodeWidth nls id) bytes num;
        write (copy "    " ("sim_" ^ name) name bytes);
        write ("  }\n");
        codeSimSample1 nls ids (num + 1);;

let codeSimSample nls buildName =
  write ("void " ^ buildName ^ "_sim_sample() {\n");
  write "  int changed = 0;\n";
  codeSimSample1 nls (portList nls) 0;
  write "  sim_count++;\n";
  write "}\n\n";;



(* FileName.c and FileName.H code generator. *)

let code nls =
  let fileName = getFile nls in
  let buildName = getBuild nls in
  let docs = getHeader nls in
  let nodes = orderNodes nls in

  NlsCode.openFile (fileName ^ ".h");
  NlsCode.codeDocs nls "//" docs buildName;
  write (codeHeader nls buildName);
  NlsCode.closeFile ();

  NlsCode.openFile (fileName ^ ".c");
  NlsCode.codeDocs nls "//" docs buildName;
  if not (Lp.embeddedC ()) then begin
    write "#include <stdio.h>\n";
    write "#include <stdlib.h>\n";
    write "#include <time.h>\n";
  end;
  write ("#include \"" ^ fileName ^ ".h\"\n");
  write (Lp.externalC () ^ "\n");
  codeDecl nls buildName nodes;
  codePorts nls buildName;
  codeCalc nls buildName nodes;
  codeInit nls buildName nodes;
  codeCycles nls buildName;
  if not (Lp.embeddedC ()) then begin
    codeSimLocals nls;
    codeSimInit nls buildName;
    codeSimEnd buildName;
    codeSimSample nls buildName;
  end;
  NlsCode.closeFile ()
;;



(* FileName_bench.c generator. *)

let rec codePortArgs nls ids =
  match ids with
    [] -> ""
  | [id] -> NlsCode.signalName nls id
  | id :: ids -> NlsCode.signalName nls id ^ ", " ^ codePortArgs nls ids;;

let codeTestBench nls cycles =
  let fileName = getFile nls in
  let buildName = getBuild nls in
  let docs = getHeader nls in
  let clocks = Nls.getClocks nls in
  let ports = Nls.portList nls in
  if List.length clocks <> 1 then raise (Error "TestBench generation requires single clock design.");
  let clock = List.hd clocks in
  NlsCode.openFile (fileName ^ "_bench.c");
  NlsCode.codeDocs nls "//" docs buildName;
  write ("#include \"" ^ fileName ^ ".h\"\n\n");
  write ("int main() {\n");
  write ("  int cycle_count;\n");
  List.iter (fun id -> write ("  unsigned char " ^ NlsCode.signalName nls id ^ "[" ^ string_of_int (nodeBytes nls id) ^ "];\n")) ports;
  write ("  " ^ buildName ^ "_ports(" ^ codePortArgs nls ports ^ ");\n");
  write ("  " ^ buildName ^ "_sim_init(\"" ^ fileName ^ "_c.vcd\");\n");
  write ("  for (cycle_count = " ^ string_of_int cycles ^ "; cycle_count > 0; cycle_count--) {\n");
  write ("    " ^ buildName ^ "_cycle_" ^ clock ^ "();\n");
  write ("    " ^ buildName ^ "_calc();\n");
  write ("    " ^ buildName ^ "_sim_sample();\n");
  write ("  }\n");
  write ("  " ^ buildName ^ "_sim_end();\n");
  write ("  return 0;\n");
  write ("}\n\n");
  NlsCode.closeFile ()
;;


(* FileName_sim.c generator. *)

let codeSimulator nls =
  let fileName = getFile nls in
  let buildName = getBuild nls in
  let docs = getHeader nls in
  let clocks = Nls.getClocks nls in
  let ports = Nls.portList nls in

  let rec cycleClock clocks =
    match clocks with
    | [] -> write ("invalid_token(tok, buf0); break;\n")
    | clock :: clocks ->
        write ("if (!strcmp(buf0, \"" ^ clock ^ "_c\")) " ^ buildName ^ "_cycle_" ^ clock ^ "(); else ");
        cycleClock clocks
  in

  let rec setPort ports =
    match ports with
    | [] -> write ("invalid_token(tok, buf0); break;\n")
    | port :: ports ->
        let name = NlsCode.signalName nls port in
        let bytes = nodeBytes nls port in
        write ("if (!strcmp(buf0, \"" ^ name ^ "\")) set_port(" ^ name ^ ", " ^ string_of_int bytes ^ ", buf1); else ");
        setPort ports
  in

  let rec queryPort ports =
    match ports with
    | [] -> write ("invalid_token(tok, buf0); break;\n")
    | port :: ports ->
        let name = NlsCode.signalName nls port in
        let bytes = nodeBytes nls port in
        write ("if (!strcmp(buf0, \"" ^ name ^ "\")) { printf(\"");
        for i = 1 to bytes do
          write ("%02X")
        done;
        write ("\\n\"");
        for i = bytes - 1 downto 0 do
          write (", " ^ name ^ "[" ^ string_of_int i ^ "]")
        done;
        write ("); fflush(stdout); } else ");
        queryPort ports
  in


  NlsCode.openFile (fileName ^ "_sim.c");
  NlsCode.codeDocs nls "//" docs buildName;
  write ("#include <stdlib.h>\n");
  write ("#include <stdio.h>\n");
  write ("#include <ctype.h>\n");
  write ("#include <string.h>\n");
  write ("#include \"" ^ fileName ^ ".h\"\n\n");

  write ("void print_help() {\n");
  write ("  printf(\"");
  write ("\\n");
  write ("NAME\\n\\n");
  write ("  " ^ fileName ^ " - Confluence generated digital logic simulator.\\n\\n");
  write ("SYNOPSIS\\n\\n");
  write ("  " ^ fileName ^ " [-h]\\n");
  write ("  stdin  : input command language\\n");
  write ("  stdout : port query results\\n\\n");
  write ("INPUT LANGUAGE\\n\\n");
  write ("  # comment     : Comment to end of line.\\n");
  write ("  !             : Initialize model.\\n");
  write ("  $ name value  : Set port value (hex).\\n");
  write ("  ? name        : Query port value.  Port value (hex) returns on stdout.\\n");
  write ("  ~ name        : Cycle clock domain.\\n");
  write ("  @             : Perform combinatorial calculation.\\n");
  if not (Lp.embeddedC ()) then begin
    write ("  [ name        : Initialize model and start a new dump file.\\n");
    write ("  |             : Sample port values to dump file.\\n");
    write ("  ]             : Close current dump file.\\n");
  end;
  write ("  EOF           : Ends simulation.\\n");
  write ("\\n");
  write ("DETAILS AND INTERFACE\\n\\n");
  NlsCode.codePrintfDocs nls docs buildName;
  write ("\\n");
  write ("\");\n");
  write ("  exit(0);\n");
  write ("}\n\n");

  if not (Lp.embeddedC ()) then begin
    write ("typedef enum {NAMEVAL, INIT, SET, QUERY, CYCLE, EVAL, OPEN, SAMPLE, CLOSE, END} token;\n\n");
  end else begin
    write ("typedef enum {NAMEVAL, INIT, SET, QUERY, CYCLE, EVAL, END} token;\n\n");
  end;

  write ("void invalid_character(int ch) {\n");
  write ("  fprintf(stderr, \"\\nInvalid character: '%c'\\n\\n\", ch);\n");
  write ("  exit(1);\n");
  write ("}\n\n");

  write ("token next_token(char *buffer) {\n");
  write ("  static int ch = 0;\n");
  write ("  if (!ch)\n");
  write ("    ch = getchar();\n");
  write ("  while (ch == '#' || isspace(ch)) {\n");
  write ("    if (ch == '#')\n");
  write ("      while (ch != '\\n')\n");
  write ("        ch = getchar();\n");
  write ("    ch = getchar();\n");
  write ("  }\n");
  write ("  if (ch == '!') { ch = 0; return INIT   ; }\n");
  write ("  if (ch == '$') { ch = 0; return SET    ; }\n");
  write ("  if (ch == '?') { ch = 0; return QUERY  ; }\n");
  write ("  if (ch == '~') { ch = 0; return CYCLE  ; }\n");
  write ("  if (ch == '@') { ch = 0; return EVAL   ; }\n");
  if not (Lp.embeddedC ()) then begin
    write ("  if (ch == '[') { ch = 0; return OPEN   ; }\n");
    write ("  if (ch == '|') { ch = 0; return SAMPLE ; }\n");
    write ("  if (ch == ']') { ch = 0; return CLOSE  ; }\n");
  end;
  write ("  if (ch == EOF) { ch = 0; return END    ; }\n");
  write ("  if (isalnum(ch) || ch == '.' || ch == '/') {\n");
  write ("    while (isalnum(ch) || ch == '.' || ch == '/' || ch == '_') {\n");
  write ("      *(buffer++) = ch;\n");
  write ("      ch = getchar();\n");
  write ("    }\n");
  write ("    *buffer = 0;\n");
  write ("    return NAMEVAL;\n");
  write ("  }\n");  
  write ("  invalid_character(ch);\n");
  write ("  exit(1);\n");
  write ("}\n\n");

  write ("void invalid_token(token t, char *buffer) {\n");
  write ("  switch (t) {\n");
  write ("    case NAMEVAL : printf(\"\\nInvalid token: %s\\n\\n\", buffer); break;\n");
  write ("    case INIT    : printf(\"\\nInvalid token: !\\n\\n\"); break;\n");
  write ("    case SET     : printf(\"\\nInvalid token: $\\n\\n\"); break;\n");
  write ("    case QUERY   : printf(\"\\nInvalid token: ?\\n\\n\"); break;\n");
  write ("    case CYCLE   : printf(\"\\nInvalid token: ~\\n\\n\"); break;\n");
  write ("    case EVAL    : printf(\"\\nInvalid token: @\\n\\n\"); break;\n");
  if not (Lp.embeddedC ()) then begin
    write ("    case OPEN    : printf(\"\\nInvalid token: [\\n\\n\"); break;\n");
    write ("    case SAMPLE  : printf(\"\\nInvalid token: |\\n\\n\"); break;\n");
    write ("    case CLOSE   : printf(\"\\nInvalid token: ]\\n\\n\"); break;\n");
  end;
  write ("    case END     : printf(\"\\nInvalid token: EOF\\n\\n\"); break;\n");
  write ("  }\n");
  write ("  exit(1);\n");
  write ("}\n\n");

  write ("char char_of_hex(int ch) {\n");
  write ("  switch (toupper(ch)) {\n");
  write ("    case '0' : return 0;\n");
  write ("    case '1' : return 1;\n");
  write ("    case '2' : return 2;\n");
  write ("    case '3' : return 3;\n");
  write ("    case '4' : return 4;\n");
  write ("    case '5' : return 5;\n");
  write ("    case '6' : return 6;\n");
  write ("    case '7' : return 7;\n");
  write ("    case '8' : return 8;\n");
  write ("    case '9' : return 9;\n");
  write ("    case 'A' : return 10;\n");
  write ("    case 'B' : return 11;\n");
  write ("    case 'C' : return 12;\n");
  write ("    case 'D' : return 13;\n");
  write ("    case 'E' : return 14;\n");
  write ("    case 'F' : return 15;\n");
  write ("    default  : invalid_character(ch); return 0;\n");
  write ("  }\n");
  write ("}\n\n");

  write ("void set_port(char *port, int bytes, char *value) {\n");
  write ("  int byte = 0;\n");
  write ("  int i = strlen(value) - 1;\n");
  write ("  while (byte < bytes) {\n");
  write ("    if (i >= 1) {\n");
  write ("      port[byte++] = (char_of_hex(value[i-1]) << 4) | char_of_hex(value[i]);\n");
  write ("      i -= 2;\n");
  write ("    }\n");
  write ("    else if (i == 0) {\n");
  write ("      port[byte++] = char_of_hex(value[i]);\n");
  write ("      i -= 1;\n");
  write ("    }\n");
  write ("    else\n");
  write ("      port[byte++] = 0;\n");
  write ("  }\n");
  write ("}\n\n");
  
  List.iter (fun id ->
    let bytes = nodeBytes nls id in
    write ("static unsigned char " ^ NlsCode.signalName nls id ^ "[" ^ string_of_int bytes ^ "] = {");
    for i = 1 to bytes do
      write ("0" ^ (if i = bytes then "" else ", "))
    done;
    write ("};\n")) ports;
  write ("\n");

  write ("int main(int argc, char **argv) {\n");
  write ("  token tok;\n");
  write ("  char  buf0[400];\n");
  write ("  char  buf1[400];\n");
  write ("  if (argc > 1)\n");
  write ("      print_help();\n");
  write ("  " ^ buildName ^ "_ports(" ^ codePortArgs nls ports ^ ");\n");
  write ("  tok = next_token(buf0);\n");
  write ("  while (tok != END) {\n");
  write ("    switch (tok) {\n");
  write ("      case INIT    : " ^ buildName ^ "_init(); break;\n");
  write ("      case SET     :\n");
  write ("        tok = next_token(buf0);\n");
  write ("        switch (tok) {\n");
  write ("          case NAMEVAL :\n");
  write ("            tok = next_token(buf1);\n");
  write ("            switch (tok) {\n");
  write ("              case NAMEVAL : "); setPort ports;
  write ("              default      : invalid_token(tok, buf0); break;\n");
  write ("            }\n");
  write ("            break;\n");
  write ("          default      : invalid_token(tok, buf0); break;\n");
  write ("        }\n");
  write ("        break;\n");
  write ("      case QUERY   :\n");
  write ("        tok = next_token(buf0);\n");
  write ("        switch (tok) {\n");
  write ("          case NAMEVAL : "); queryPort ports;
  write ("          default      : invalid_token(tok, buf0); break;\n");
  write ("        }\n");
  write ("        break;\n");
  write ("      case CYCLE   : ");
  write ("        tok = next_token(buf0);\n");
  write ("        switch (tok) {\n");
  write ("          case NAMEVAL : "); cycleClock clocks;
  write ("          default      : invalid_token(tok, buf0); break;\n");
  write ("        }\n");
  write ("        break;\n");
  write ("      case EVAL    : " ^ buildName ^ "_calc(); break;\n");
  if not (Lp.embeddedC ()) then begin
    write ("      case OPEN    :\n");
    write ("        tok = next_token(buf0);\n");
    write ("        switch (tok) {\n");
    write ("          case NAMEVAL : " ^ buildName ^ "_sim_init(buf0); break;\n");
    write ("          default      : invalid_token(tok, buf0); break;\n");
    write ("        }\n");
    write ("        break;\n");
    write ("      case SAMPLE  : " ^ buildName ^ "_sim_sample(); break;\n");
    write ("      case CLOSE   : " ^ buildName ^ "_sim_end(); break;\n");
  end;
  write ("      default      : invalid_token(tok, buf0); break;\n");
  write ("    }\n");
  write ("    tok = next_token(buf0);\n");
  write ("  }\n");
  write ("  return 0;\n");
  write ("}\n\n");

  NlsCode.closeFile ()
;;

