(*
    Confluence System Design Language Compiler
    Copyright (C) 2003-2004 Tom Hawkins (tomahawkins@yahoo.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*)


(** File writer. *)

let file = ref stdout;;

(** Open a new file. *)
let openFile fileName =
  file := open_out fileName
;;

(** Close current file. *)
let closeFile () =
  close_out !file
;;

(** Write string to file. *)
let write str =
  output_string !file str
;;


(** Documentation functions. *)

let rec buildInfoClocks clocks =
  match clocks with
    [] -> ""
  | clock :: clocks -> clock ^ "_c  " ^ buildInfoClocks clocks;;

let rec buildInfoPorts nls pids =
  match pids with
    [] -> ""
  | pid :: pids ->
      begin
        match Nls.node nls pid with
          Lp.VecInput    (name, _, width) -> "  Vector Input  : " ^ name ^ "_i(" ^ string_of_int width ^ ")\n"
        | Lp.VecOutput   (name, _, width) -> "  Vector Output : " ^ name ^ "_o(" ^ string_of_int width ^ ")\n"
        | Lp.VecInOut    (name, _, width) -> "  Vector InOut  : " ^ name ^ "_b(" ^ string_of_int width ^ ")\n"
        | _ -> raise (Nls.Error "NlsCode.codeDocs:  Requires ports.")
      end ^ buildInfoPorts nls pids;;

let buildInfo nls buildName =
  "\n\nGenerated by the Confluence System Design Language " ^ Version.version ^ "  --  http://www.launchbird.com/\n\n" ^
  "Build Date : " ^ Report.date () ^ "\n\n" ^
  "Interface\n\n" ^
  "  Build Name    : " ^ buildName ^ "\n" ^
  "  Clock Domains : " ^ buildInfoClocks (Nls.getClocks nls) ^ "\n" ^
                         buildInfoPorts nls (Nls.portList nls) ^ "\n\n"
;;

let codeComment cmt header =
  match header with
    "" -> ()
  | _ ->
      let f ch =
        match ch with
          '\n' -> write ("\n" ^ cmt ^ "  ")
        | _    -> write (String.make 1 ch)
      in
      write (cmt ^ "  ");
      String.iter f header;
      write ("\n");
;;

let codeDocs nls cmt docs buildName =
  codeComment cmt ("\n" ^ docs ^ buildInfo nls buildName);
  write ("\n\n");
;;

let codePrintfDocs nls docs buildName =
  write (String2.replace_printf_chars ("\n" ^ docs ^ buildInfo nls buildName))
;;



(** Signal name functions. *)

let signalName nls id =
  match Nls.node nls id with
    Lp.VecInput  (name, _, _) -> name ^ "_i"
  | Lp.VecOutput (name, _, _) -> name ^ "_o"
  | Lp.VecInOut  (name, _, _) -> name ^ "_b"
  | Lp.VecAssert _ -> ""  (* No signal names used. *)
  | Lp.VecPrint  _ -> ""  (* No signal names used. *)
  | _ -> "n" ^ string_of_int (Nls.idToInt id);;

let producerSignalName nls cId cPort =
  signalName nls (Nls.source nls cId cPort);;

let topLevelSourceName nls id port =
  let sid = Nls.source nls id port in
  match Nls.node nls sid with
    Lp.VecInput (n, _, _) -> n ^ "_i"
  | Lp.VecInOut (n, _, _) -> n ^ "_b"
  | _ -> "n" ^ string_of_int (List2.index sid (Nls.IdSet.elements nls.Nls.mainSpace.Nls.spaceOutputs) 1);;


(** Misc helpers. *)

let bytes = Lp.bytes;;

let nodeBytes nls id =
  bytes (Nls.nodeWidth nls id);;

let rec byteRepeat f bytes =
  if bytes <= 0 then ""
  else
    byteRepeat f (bytes - 1) ^ f ("[" ^ string_of_int (bytes - 1) ^ "]");;
    
let copy indent sink src bytes =
  let f str =
    indent ^ sink ^ str ^ " = " ^ src ^ str ^ ";\n"
  in
    byteRepeat f bytes;;

let nameByte name byte =
  name ^ "[" ^ string_of_int byte ^ "]";;
  
let rec intsToString is =
  match is with
    [] -> ""
  | i :: [] -> string_of_int i
  | i :: is -> string_of_int i ^ ", " ^ intsToString is;;

let mask width =
  int_of_float (2.0 ** float_of_int (((width - 1) mod 8) + 1)) - 1;;

let binToInt bits =
  let rec f width bits sofar =
    if width = 0 then sofar else
      let width = width - 1
      in
        match String.get bits 0 with
          '0' -> f width (String.sub bits 1 width) sofar
        | '1' -> f width (String.sub bits 1 width) (sofar + int_of_float (2.0 ** float_of_int width))
        | _ -> raise (Nls.Error "NlsCode.binToInt:  Invalid bit string.")
  in
    f (String.length bits) bits 0;;

(** Converts binary string to hex string. *)
let binToHex bin =
  let binToHexLookUp bin =
    match bin with
      "0000" -> "0"
    | "0001" -> "1"
    | "0010" -> "2"
    | "0011" -> "3"
    | "0100" -> "4"
    | "0101" -> "5"
    | "0110" -> "6"
    | "0111" -> "7"
    | "1000" -> "8"
    | "1001" -> "9"
    | "1010" -> "A"
    | "1011" -> "B"
    | "1100" -> "C"
    | "1101" -> "D"
    | "1110" -> "E"
    | "1111" -> "F"
    | _ -> raise (Nls.Error ("NlsCode.binToHex: Invalid binary string: " ^ bin))
  in
  let rec binToHex bin =
    let l = String.length bin in
    if l = 0 then "" else
    if l <= 4 then
      binToHexLookUp (String.make (4 - l) '0' ^ bin)
    else
      binToHex (String.sub bin 0 (l - 4)) ^ binToHexLookUp (String.sub bin (l - 4) 4)
  in
  "0x" ^ binToHex bin
;;

(** Writes a sequence with a delimiter. *)
let writeSequence func items delimiter =
  let rec f items =
    match items with
      [] -> ()
    | item :: items ->
        func item;
        if items <> [] then write delimiter;
        f items
  in
  f items
;;

