(*
    Confluence System Design Language Compiler
    Copyright (C) 2003-2004 Tom Hawkins (tomahawkins@yahoo.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*)


open Nls;;

let write = NlsCode.write;;


(** Misc helpers. *)

let bitRepeat func width =
  let rec f bit =
    if bit < width then
      let bitStr = "[" ^ string_of_int bit ^ "]" in
      write (func bitStr);
      f (bit + 1)
  in
  f 0
;;

let copy sink src width =
  bitRepeat (fun str -> "    " ^ sink ^ str ^ " := " ^ src ^ str ^ ";\n") width
;;


(** Declarations. *)

let codeNodeDeclName name width =
  write ("    " ^ name ^ " : array 0 .. " ^ string_of_int (width - 1) ^ " of boolean;\n")
;;


let codeNodeDecl nls id =
  if Lp.isCoded (node nls id) then
    let name = NlsCode.signalName nls id in
    let width = nodeWidth nls id in
    match node nls id with
      Lp.VecAssert _ -> ()
    | Lp.VecPrint  _ -> ()
    
    | Lp.VecAdd _ ->
        codeNodeDeclName name width;
        write ("    " ^ name ^ "cr : array 0 .. " ^ string_of_int (width - 1) ^ " of 0 .. 3;\n");

    | Lp.VecSub _ ->
        codeNodeDeclName name width;
        write ("    " ^ name ^ "cr : array 0 .. " ^ string_of_int (width - 1) ^ " of -2 .. 1;\n");

    | Lp.VecMul _ ->
        codeNodeDeclName name width;
        write ("    " ^ name ^ "r : 0 .. " ^ string_of_int (int_of_float (2.0 ** float_of_int (width * 2)) - 1) ^ ";\n");

    | Lp.VecRom (_, _, vs) ->
        codeNodeDeclName name width;
        write ("    " ^ name ^ "m : array 0 .. " ^ string_of_int (List.length vs) ^ " of array 0 .. " ^ string_of_int width ^ " of boolean;\n")

    | Lp.VecRamMem (dw, vs) ->
        write ("    " ^ name ^ " : array 0 .. " ^ string_of_int (List.length vs) ^ " of array 0 .. " ^ string_of_int dw ^ " of boolean;\n")

    | Lp.VecRamWrite _ -> ()

    | Lp.VecRamRead (_, rbw, aw, dw) ->
        codeNodeDeclName name dw;
        if not rbw then codeNodeDeclName (name ^ "a") aw;

    | _ -> 
        codeNodeDeclName name width;
;;


(** Node function helpers. *)

(** Accesses a memory. *)
let rec memoryAccess memoryName addrName addrWidth addrCurrent addrStage addrBit dataBit =
  if addrBit = addrWidth then
    memoryName ^ "[" ^ string_of_int addrCurrent ^ "]" ^ dataBit
  else
    "(! " ^ addrName ^ "[" ^ string_of_int addrBit ^ "] & " ^ memoryAccess memoryName addrName addrWidth (addrCurrent            ) (addrStage * 2) (addrBit + 1) dataBit ^
    " | " ^ addrName ^ "[" ^ string_of_int addrBit ^ "] & " ^ memoryAccess memoryName addrName addrWidth (addrCurrent + addrStage) (addrStage * 2) (addrBit + 1) dataBit ^ ")"
;;


(** Combination and permutation. *)

let rec combine num items =
  if num = 1 then
    List.map (fun item -> [item]) items
  else
    combine1 num items

and combine1 num items =
  match items with
  | [] -> []
  | item :: items -> 
      List.map (fun sub -> item :: sub) (combine (num - 1) items) @ combine1 num items
;;

let permute items =
  let rec f num =
    if num > 0 then
      combine num items @ f (num - 1)
    else
      []
  in
  f (List.length items)
;;
      
    
    
    

(** Node function (combinatorial and sequential). *)
let codeNodeAssign nls id =
  if Lp.isCoded (node nls id) then
    let name = NlsCode.signalName nls id in
    let input n = NlsCode.producerSignalName nls id n in
    match node nls id with 
  
    | Lp.VecInput _ -> ()
  
    | Lp.VecOutput (_, _, w) ->
        copy name (input 0) w
  
    | Lp.VecInOut (n, _, w) -> raise (Error "NlsNuSmv.codeNodeCalc:  VecInOut not supported yet.") (* XXX Not supported yet. *)
    
    | Lp.VecAssert _ -> ()  (* Assertions handled separately. *)

    | Lp.VecPrint _ -> ()   (* No printing supported. *)
  
    | Lp.VecConst s ->
        let rec f str bit =
          if str <> "" then begin
            write ("    " ^ name ^ "[" ^ string_of_int bit ^ "] := " ^ String2.take_right str 1 ^ ";\n");
            f (String2.left_string str) (bit + 1)
          end
        in
        f s 0
        
    | Lp.VecBuf w ->
        copy name (input 0) w
  
    | Lp.VecNot w ->
        bitRepeat (fun bit -> "    " ^ name ^ bit ^ " := ! " ^ input(0) ^ bit ^ ";\n") w

    | Lp.VecAnd w ->
        bitRepeat (fun bit -> "    " ^ name ^ bit ^ " := " ^ input(0) ^ bit ^ " & " ^ input(1) ^ bit ^ ";\n") w
  
    | Lp.VecOr w ->
        bitRepeat (fun bit -> "    " ^ name ^ bit ^ " := " ^ input(0) ^ bit ^ " | " ^ input(1) ^ bit ^ ";\n") w
  
    | Lp.VecXor w ->
        bitRepeat (fun bit -> "    " ^ name ^ bit ^ " := (" ^ input 0 ^ bit ^ " & ! " ^ input 1 ^ bit ^ ") | (! " ^ input 0 ^ bit ^ " & " ^ input 1 ^ bit ^ ");\n") w
        (*
        bitRepeat (fun bit -> "    " ^ name ^ bit ^ " := " ^ input(0) ^ bit ^ " xor " ^ input(1) ^ bit ^ ";\n") w
        *)
  
    | Lp.VecConcat (0, 0) -> ()
    | Lp.VecConcat (w, 0) ->
        copy name (input 0) w;
    | Lp.VecConcat (0, w) ->
        copy name (input 1) w;
    | Lp.VecConcat (w0, w1) ->
        copy name (input 1) w1;
        for i = w1 to w0 + w1 - 1 do
          write ("    " ^ name ^ "[" ^ string_of_int i ^ "] := " ^ (input 0) ^ "[" ^ string_of_int (i - w1) ^ "];\n")
        done
  
    | Lp.VecSelect (_, bs) ->
        let rec f bits b =
          match bits with
            [] -> ()
          | bit :: bits ->
              write ("    " ^ name ^ "[" ^ string_of_int b ^ "] := "  ^ (input 0) ^ "[" ^ string_of_int bit ^ "];\n");
              f bits (b + 1)
        in
        f (List.rev bs) 0

    | Lp.VecEqu w ->
        write ("    " ^ name ^ "[0] := ");
        for i = 0 to w - 1 do
          write ("(" ^ input 0 ^ "[" ^ string_of_int i ^ "] <-> " ^ input 1 ^ "[" ^ string_of_int i ^ "])");
          if i <> w - 1 then
            write (" & ");
        done;
        write (";\n")
  
    | Lp.VecAdd w ->
        write ("    " ^ name ^ "cr[0] := " ^ input 0 ^ "[0] + " ^ input 1 ^ "[0];\n");
        write ("    " ^ name ^   "[0] := " ^ name ^ "cr[0] mod 2;\n");
        for i = 1 to w - 1 do
          let b0 = "[" ^ string_of_int i ^ "]" in
          let b1 = "[" ^ string_of_int (i - 1) ^ "]" in
          write ("    " ^ name ^ "cr" ^ b0 ^ " := " ^ input 0 ^ b0 ^ " + " ^ input 1 ^ b0 ^ " + (" ^ name ^ "cr" ^ b1 ^ " / 2);\n");
          write ("    " ^ name ^        b0 ^ " := " ^ name ^ "cr" ^ b0 ^ " mod 2;\n");
        done
  
    | Lp.VecSub w ->
        write ("    " ^ name ^ "cr[0] := " ^ input 0 ^ "[0] - " ^ input 1 ^ "[0];\n");
        write ("    " ^ name ^   "[0] := " ^ name ^ "cr[0] mod 2;\n");
        for i = 1 to w - 1 do
          let b0 = "[" ^ string_of_int i ^ "]" in
          let b1 = "[" ^ string_of_int (i - 1) ^ "]" in
          write ("    " ^ name ^ "cr" ^ b0 ^ " := " ^ input 0 ^ b0 ^ " - " ^ input 1 ^ b0 ^ " + (" ^ name ^ "cr" ^ b1 ^ " / 2);\n");
          write ("    " ^ name ^        b0 ^ " := " ^ name ^ "cr" ^ b0 ^ " mod 2;\n");
        done
  
    | Lp.VecMul w ->
        let rec f0 n bit =
          if bit < w then begin
            write (" + (" ^ n ^ "[" ^ string_of_int bit ^ "] * " ^ string_of_int (int_of_float (2.0 ** float_of_int bit)) ^ ")");
            f0 n (bit + 1)
          end
        in
        let rec f1 bit =
          if bit < w then
            let shift = string_of_int (int_of_float (2.0 ** float_of_int bit)) in
            write ("    " ^ name ^ "[" ^ string_of_int bit ^ "] := " ^ (if shift = "1" then name ^ "r" else "(" ^ name ^ "r / " ^ shift ^ ")") ^ " mod 2;\n");
            f1 (bit + 1)
        in
        write ("    " ^ name ^ "r := (" ^ input 0 ^ "[0]");
        f0 (input 0) 1;
        write (") * (" ^ input 1 ^ "[0]");
        f0 (input 1) 1;
        write (");\n");
        f1 0
        
    | Lp.VecSwitch w ->
        bitRepeat (fun bit ->
          "    " ^ name ^ bit ^ " := case " ^ input 0 ^ "[0] : " ^ input 1 ^ bit ^ ";" ^
                                       " !" ^ input 0 ^ "[0] : " ^ input 2 ^ bit ^ "; esac;\n"
        ) w

    | Lp.VecExtComb _ | Lp.VecExtSequ _ ->
        raise (Nls.Error "NuSMV code generation does not support external components.")

    | Lp.VecReg (clock, w) ->
        bitRepeat (fun bit ->
          "    init(" ^ name ^ bit ^ ") := 0;\n" ^
          "    next(" ^ name ^ bit ^ ") := " ^ clock ^ "_c_re & (" ^
          "! " ^ input 0 ^ "[0] & (" ^ input 1 ^ "[0] & " ^ input 2 ^ bit ^ " | ! " ^ input 1 ^ "[0]" ^ " & " ^ name ^ bit ^ ")" ^
          ") | ! " ^ clock ^ "_c_re & " ^ name ^ bit ^ ";\n"
        ) w

    | Lp.VecRom (clock, w, vs) ->
        let rec word addr bit value =
          if bit < w then begin
            write ("    " ^ name ^ "m[" ^ string_of_int addr ^ "][" ^ string_of_int bit ^ "] := " ^ Intbig.string_of_intbig (Intbig.modu value Intbig.two) ^ ";\n");
            word addr (bit + 1) (Intbig.div value Intbig.two)
          end
        in
        let rec rom addr values =
          match values with
          | [] -> ()
          | value :: values -> word addr 0 value; rom (addr + 1) values
        in
        rom 0 vs;
        let addrWidth = Lp.bits (List.length vs) in
        bitRepeat (fun bit ->
          "    init(" ^ name ^ bit ^ ") := 0;\n" ^    
          "    next(" ^ name ^ bit ^ ") := case (" ^ clock ^ "_c_re & " ^ input 0 ^ "[0]) : " ^ memoryAccess (name ^ "m") (input 1) addrWidth 0 1 0 bit ^ ";" ^
                                             " !(" ^ clock ^ "_c_re & " ^ input 0 ^ "[0]) : " ^ name ^ bit ^ "; esac;\n"
        ) w

    | Lp.VecRamMem (dw, vs) ->
        let depth = List.length vs in
        let aw = Lp.bits depth in
        let ports = IdPortSet.fold (fun idPort portIds ->
          let pid = fst (idPortToPair idPort) in
          (match node nls pid with
          | Lp.VecRamWrite (c, _, _) ->
              (c ^ "_c_re", NlsCode.producerSignalName nls pid 1,
                            NlsCode.producerSignalName nls pid 2,
                            NlsCode.producerSignalName nls pid 3) :: portIds
          | Lp.VecRamRead _ -> portIds
          | _ -> raise (Error "NlsNuSmv.codeNodeAssign: Expecting VecRamPort."))) (sinks nls id) []
        in
        let rec addrCompare addrName addrBit addr =
          if addrBit >= aw then
            ""
          else
            addrCompare addrName (addrBit + 1) (addr / 2) ^ (if addrBit < aw - 1 then " & " else "") ^
            (if addr mod 2 = 1 then "" else "! ") ^  addrName ^ "[" ^ string_of_int addrBit ^ "]"
        in
        let writeOption ports addr dataBit =
          write "        ";
          NlsCode.writeSequence (function (clock, write_enable, write_addr, _) ->
            write ("(" ^ clock ^ " & " ^ write_enable ^ "[0] & (" ^ addrCompare write_addr 0 addr ^ "))")) ports " & ";
          write " : ";
          NlsCode.writeSequence (function (_, _, _, write_data) ->
            write (write_data ^ "[" ^ string_of_int dataBit ^ "]")) ports " union ";
          write ";\n"
        in
        for a = 0 to depth - 1 do
          for d = 0 to dw - 1 do
            write ("    init(" ^ name ^ "[" ^ string_of_int a ^ "][" ^ string_of_int d ^ "]) := " ^ Intbig.string_of_intbig (Intbig.modu (Intbig.shift_right (List.nth vs a) (Intbig.intbig_of_int d)) Intbig.two) ^ ";\n");
            write ("    next(" ^ name ^ "[" ^ string_of_int a ^ "][" ^ string_of_int d ^ "]) :=\n");
            write ("      case\n");
            List.iter (fun ports -> writeOption ports a d) (permute ports);
            write ("        1 : " ^ name ^ "[" ^ string_of_int a ^ "][" ^ string_of_int d ^ "];\n");
            write ("      esac;\n");
          done;
        done;

    | Lp.VecRamWrite _ -> ()

    | Lp.VecRamRead (clock, rbw, aw, dw) ->
        if rbw then
          bitRepeat (fun bit ->
            "    init(" ^ name ^ bit ^ ") := 0;\n" ^
            "    next(" ^ name ^ bit ^ ") := case (" ^ clock ^ "_c_re & " ^ input 1 ^ "[0]) : " ^ memoryAccess (input 0) (input 4) aw 0 1 0 bit ^ ";" ^
                                               " !(" ^ clock ^ "_c_re & " ^ input 1 ^ "[0]) : " ^ name ^ bit ^ "; esac;\n" 
          ) dw
        else begin
          bitRepeat (fun bit ->
            "    init(" ^ name ^ "a" ^ bit ^ ") := 0;\n" ^
            "    next(" ^ name ^ "a" ^ bit ^ ") := case (" ^ clock ^ "_c_re & " ^ input 1 ^ "[0]) : " ^ input 4 ^ bit ^ ";" ^
                                                     " !(" ^ clock ^ "_c_re & " ^ input 1 ^ "[0]) : " ^ name ^ "a" ^ bit ^ "; esac;\n"
          ) aw;
          bitRepeat (fun bit ->
            "    " ^ name ^ bit ^ " := " ^ memoryAccess (input 0) (name ^ "a") aw 0 1 0 bit ^ ";\n" 
          ) dw
        end

;;


let codeNodeSpec nls id =
  if Lp.isCoded (node nls id) then
    let name = NlsCode.signalName nls id in
    let input n = NlsCode.producerSignalName nls id n in
    match node nls id with
    | Lp.VecAssert msg ->
        NlsCode.codeComment "--  " ("Assertion Violation\n\n" ^ msg);
        write ("  SPEC AG " ^ input 0 ^ "[0]\n");
    | _ -> ()
;;


(** FileName.smv code generator. *)

let code nls =
  let fileName = getFile nls in
  let buildName = getBuild nls in
  let docs = getHeader nls in
  let nodes = orderNodes nls in
  let clocks = List.map (fun clock -> clock ^ "_c") (getClocks nls) in
  NlsCode.openFile (fileName ^ ".smv");
  NlsCode.codeDocs nls "--" docs buildName;
  
  write ("MODULE " ^ buildName ^ "\n");
  write ("  VAR\n");
  List.iter (fun clock ->
    write ("    " ^ clock ^ " : boolean;\n");
    write ("    " ^ clock ^ "_last_value : boolean;\n");
    write ("    " ^ clock ^ "_re : boolean;\n")
  ) clocks;
  List.iter (fun id -> codeNodeDecl nls id) nodes;
  write ("  ASSIGN\n");
  List.iter (fun clock ->
    write ("    init(" ^ clock ^ "_last_value) := 0;\n");
    write ("    next(" ^ clock ^ "_last_value) := " ^ clock ^ ";\n");
    write ("    " ^ clock ^ "_re := " ^ clock ^ " & ! " ^ clock ^ "_last_value;\n");
  ) clocks;
  List.iter (fun id -> codeNodeAssign nls id) nodes;
  List.iter (fun id -> codeNodeSpec nls id) nodes;
  write ("\n");
  write ("MODULE main\n");
  write ("  VAR\n");
  write ("    " ^ buildName ^ " : " ^ buildName ^ ";\n");
  write ("\n\n");

  NlsCode.closeFile ()
;;

(** Unit tests. *)

let printIntListList ls =
  print_string "[";
  List.iter (fun l ->
    print_string "[";
    List.iter (fun i -> print_string (string_of_int i ^ "; ")) l;
    print_string "]") ls;
  print_string "]\n"
;;

let test n =
  Ut.assert_bool n "combine3.1" (combine 1 [0; 1; 2] = [[0]; [1]; [2]]);
  Ut.assert_bool n "combine3.2" (combine 2 [0; 1; 2] = [[0; 1]; [0; 2]; [1; 2]]);
  Ut.assert_bool n "combine3.3" (combine 3 [0; 1; 2] = [[0; 1; 2]]);
  Ut.assert_bool n "combine4.1" (combine 1 [0; 1; 2; 3] = [[0]; [1]; [2]; [3]]);
  Ut.assert_bool n "combine4.2" (combine 2 [0; 1; 2; 3] = [[0; 1]; [0; 2]; [0; 3]; [1; 2]; [1; 3]; [2; 3]]);
  Ut.assert_bool n "combine4.3" (combine 3 [0; 1; 2; 3] = [[0; 1; 2]; [0; 1; 3]; [0; 2; 3]; [1; 2; 3]]);
  Ut.assert_bool n "combine4.4" (combine 4 [0; 1; 2; 3] = [[0; 1; 2; 3]]);

  Ut.assert_bool n "permute1" (permute [0] = [[0]]);
  Ut.assert_bool n "permute2" (permute [0; 1] = [[0; 1]; [0]; [1]]);
  Ut.assert_bool n "permute3" (permute [0; 1; 2] = [[0; 1; 2]; [0; 1]; [0; 2]; [1; 2]; [0]; [1]; [2]]);
;;


