(*
    Confluence System Design Language Compiler
    Copyright (C) 2003-2004 Tom Hawkins (tomahawkins@yahoo.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*)


open Nls;;

(* Basic optimization functions. *)

(* Remove null effect logic. *)

let bypassBuffers nls =
  let rec connectToNonBuffer cId cPort pId =
    match node nls pId with
      Lp.VecBuf _ -> connectToNonBuffer cId cPort (source nls pId 0)
    | _ -> reconnect nls cId cPort pId
  in
  let rec f1 cId cPort cArity =
    if cPort < cArity then begin
      connectToNonBuffer cId cPort (source nls cId cPort);
      f1 cId (cPort + 1) cArity
    end
  in
  let f2 id node =
    match node with
      Lp.VecInput _ -> ()
    | Lp.VecConst _ -> ()
    | Lp.VecBuf   _ -> ()
    | _ -> f1 id 0 (nodeArity nls id)
  in
  Hashtbl.iter f2 nls.nodes;;

let rec nodesRequired nls ids sofar =
  if IdSet.is_empty ids then
    sofar
  else
    let id = IdSet.choose ids in
    nodesRequired nls (IdSet.union (IdSet.remove id ids) (IdSet.diff (required nls id) sofar)) (IdSet.add id sofar)
;;

let interfaceNodes nls =
  let f interfaceSet id =
    match node nls id with
    | Lp.VecInput  _
    | Lp.VecOutput _
    | Lp.VecInOut  _
    | Lp.VecAssert _
    | Lp.VecPrint  _ ->
        IdSet.add id interfaceSet
    | _ ->
        interfaceSet
  in
  List.fold_left f IdSet.empty (nodeList nls)
;;


let removeNullEffect nls =
  bypassBuffers nls;
  let ids = nodesRequired nls (interfaceNodes nls) IdSet.empty in
  let f1 id node =
    match node with
    | Lp.VecInput    _
    | Lp.VecAssert   _
    | Lp.VecPrint    _
    | Lp.VecRamWrite _ -> ()
    | _ -> if not (IdSet.mem id ids) then delete nls id
  in
  Hashtbl.iter f1 nls.nodes
;;


(* Constant propagation. *)

type signal = Variable of id | Constant of id * string;;


let binNot a =
  String2.map (fun a ->
    match a with
      '1' -> '0'
    | '0' -> '1'
    | _ -> raise (Error "NlsOpt.binNot:  Invalid character.")) a;;

let binAnd a b =
  String2.map2 (fun a b ->
    match (a, b) with
      ('0', '0') -> '0'
    | ('0', '1') -> '0'
    | ('1', '0') -> '0'
    | ('1', '1') -> '1'
    | _ -> raise (Error "NlsOpt.binAnd:  Invalid character.")) a b;;

let binAdd a b =
  let inputs = List.combine (String2.chars_of_string a) (String2.chars_of_string b) in
  let (_, r) = List.fold_right (fun (a, b) (c, r) ->
    match (a, b, c) with
    | ('0', '0', '0') -> '0', "0" ^ r
    | ('0', '0', '1') -> '0', "1" ^ r
    | ('0', '1', '0') -> '0', "1" ^ r
    | ('0', '1', '1') -> '1', "0" ^ r
    | ('1', '0', '0') -> '0', "1" ^ r
    | ('1', '0', '1') -> '1', "0" ^ r
    | ('1', '1', '0') -> '1', "0" ^ r
    | ('1', '1', '1') -> '1', "1" ^ r
    | _ -> raise (Error "NlsOpt.binAdd:  Invalid character.")) inputs ('0', "") in
  r
;;

let binSub a b =
  let inputs = List.combine (String2.chars_of_string a) (String2.chars_of_string b) in
  let (_, r) = List.fold_right (fun (a, b) (c, r) ->
    match (a, b, c) with
    | ('0', '0', '0') -> '0', "0" ^ r
    | ('0', '0', '1') -> '1', "1" ^ r
    | ('0', '1', '0') -> '1', "1" ^ r
    | ('0', '1', '1') -> '1', "0" ^ r
    | ('1', '0', '0') -> '0', "1" ^ r
    | ('1', '0', '1') -> '0', "0" ^ r
    | ('1', '1', '0') -> '0', "0" ^ r
    | ('1', '1', '1') -> '1', "1" ^ r
    | _ -> raise (Error "NlsOpt.binSub:  Invalid character.")) inputs ('0', "") in
  r
;;

let binXor a b =
  String2.map2 (fun a b ->
    match (a, b) with
      ('0', '0') -> '0'
    | ('0', '1') -> '1'
    | ('1', '0') -> '1'
    | ('1', '1') -> '0'
    | _ -> raise (Error "NlsOpt.binXor:  Invalid character.")) a b;;

let binOr a b =
  String2.map2 (fun a b ->
    match (a, b) with
      ('0', '0') -> '0'
    | ('0', '1') -> '1'
    | ('1', '0') -> '1'
    | ('1', '1') -> '1'
    | _ -> raise (Error "NlsOpt.binOr:  Invalid character.")) a b;;

let binAll bin char =
  bin = String.make (String.length bin) char;;

let rec binSelect bin bits =
  match bits with
    [] -> ""
  | bit :: bits -> String.make 1 bin.[String.length bin - 1 - bit] ^ binSelect bin bits;;

let constantPropagation nls =
  let queue = Queue.create () in

  let addConstConsumers id = 
    IdPortSet.iter (fun idPort -> match idPortToPair idPort with (id, _) -> Queue.add id queue) (sinks nls id)
  in

  let inputs id =
    List.map (fun sid -> match node nls sid with Lp.VecConst s -> Constant (sid, s) | _ -> Variable sid) (sourceList nls id)
  in

  let rec processNodes () =
    if not (Queue.is_empty queue) then begin
      let id  = Queue.take queue in
      if contains nls id then begin
        let space = Hashtbl.find nls.n2s id in
        let _ = node nls id in
        let _ = inputs id in
        match (node nls id, inputs id) with

          (Lp.VecBuf _, [Constant (s, v)]) -> addConstConsumers id; redirect nls id s;

        | (Lp.VecNot _, [Constant (s, v)]) -> addConstConsumers id; redirect nls id (newVecConst nls space [] (binNot v));

        | (Lp.VecAnd _, [Constant (s0, v0); Variable s1])
        | (Lp.VecAnd _, [Variable s1; Constant (s0, v0)]) ->
            if binAll v0 '0' then begin
              addConstConsumers id;
              redirect nls id s0
            end else if binAll v0 '1' then begin
              redirect nls id s1
            end
        | (Lp.VecAnd _, [Constant (s0, v0); Constant (s1, v1)]) ->
            addConstConsumers id;
            redirect nls id (newVecConst nls space [] (binAnd v0 v1))
        
        | (Lp.VecAdd _, [Constant (s0, v0); Variable s1])
        | (Lp.VecAdd _, [Variable s1; Constant (s0, v0)]) ->
            if binAll v0 '0' then redirect nls id s1
        | (Lp.VecAdd _, [Constant (s0, v0); Constant (s1, v1)]) ->
            addConstConsumers id;
            redirect nls id (newVecConst nls space [] (binAdd v0 v1))
        
        | (Lp.VecSub _, [Variable s1; Constant (s0, v0)]) ->  (* XXX Could check [c; v] where c is all 1s. *)
            if binAll v0 '0' then redirect nls id s1
        | (Lp.VecSub _, [Constant (s0, v0); Constant (s1, v1)]) ->
            addConstConsumers id;
            redirect nls id (newVecConst nls space [] (binSub v0 v1))
        

        | (Lp.VecOr _, [Constant (s0, v0); Variable s1])
        | (Lp.VecOr _, [Variable s1; Constant (s0, v0)]) ->
            if binAll v0 '1' then begin
              addConstConsumers id;
              redirect nls id s0
            end else if binAll v0 '0' then begin
              redirect nls id s1
            end
        | (Lp.VecOr _, [Constant (s0, v0); Constant (s1, v1)]) ->
            addConstConsumers id;
            redirect nls id (newVecConst nls space [] (binOr v0 v1))

        | (Lp.VecXor w, [Constant (s0, v0); Variable s1])
        | (Lp.VecXor w, [Variable s1; Constant (s0, v0)]) ->
            if binAll v0 '0' then begin
              redirect nls id s1
            end else if binAll v0 '1' then begin
              redirect nls id (newVecNot nls space [] s1)
            end
        | (Lp.VecXor _, [Constant (s0, v0); Constant (s1, v1)]) ->
            addConstConsumers id;
            redirect nls id (newVecConst nls space [] (binXor v0 v1))

        | (Lp.VecConcat (w, _), [Constant (s0, v0); Variable s1]) 
        | (Lp.VecConcat (_, w), [Variable s1; Constant (s0, v0)]) ->
            if w = 0 then
              redirect nls id s1
        | (Lp.VecConcat _, [Constant (_, v0); Constant (_, v1)]) ->
            addConstConsumers id;
            redirect nls id (newVecConst nls space [] (v0 ^ v1))

        | (Lp.VecSelect (_, bits), [Constant (s0, v0)]) ->
            addConstConsumers id;
            redirect nls id (newVecConst nls space [] (binSelect v0 bits))

        | (Lp.VecEqu _, [Constant (_, v0); Constant (_, v1)]) ->
            addConstConsumers id;
            redirect nls id (newVecConst nls space [] (if v0 = v1 then "1" else "0"))

        | (Lp.VecSwitch _, [_; Constant (_, v1); Constant (_, v2)]) ->
            if v1 = v2 then begin
              addConstConsumers id;
              redirect nls id (newVecConst nls space [] v1)
            end;
        | (Lp.VecSwitch _, [Constant (_, v0); Variable s1; Variable s2])
        | (Lp.VecSwitch _, [Constant (_, v0); Variable s1; Constant (s2, _)])
        | (Lp.VecSwitch _, [Constant (_, v0); Constant (s1, _); Variable s2]) ->
            addConstConsumers id;
            redirect nls id (if v0 = "1" then s1 else s2)

        (* XXX More left to do. *)

        | _ -> ()
      end;
      processNodes ();
    end
  in

  (* Add initial nodes to queue. *)
  IdSet.iter (fun id -> addConstConsumers id) nls.constants;
  processNodes ();
  
;;


(* Basic validity checking. *)

let checkConnections nls =
  let rec f2 widths cId cPort =
    match widths with
      [] -> ()
    | width :: widths ->
        begin
          try
            if nodeWidth nls (source nls cId cPort) <> width then
              raise (Error ("Connection width mismatch.  Consumer: n" ^ string_of_int (idToInt cId) ^ "  " ^ nodeType nls cId ^ string_of_int cPort ^ "(" ^ string_of_int width ^ ")  Producer: n" ^
                            string_of_int (idToInt (source nls cId cPort)) ^ "  " ^ nodeType nls (source nls cId cPort) ^ "(" ^ string_of_int (nodeWidth nls (source nls cId cPort)) ^ ")"));
          with Not_found -> raise (Error ("Required node input unconnected.  Type: " ^ nodeType nls cId ^ "  Port: " ^ string_of_int cPort));
        end;
        f2 widths cId (cPort + 1)
  in
    let f1 id node =
      f2 (nodeInputWidths nls id) id 0;
    in
      Hashtbl.iter f1 nls.nodes;;
    


(* Unit tests. *)

let test n =
  Ut.assert_bool n "binNot" (binNot "1100" = "0011");
  Ut.assert_bool n "binAnd" (binAnd "1100" "1010" = "1000");
  Ut.assert_bool n "binXor" (binXor "1100" "1010" = "0110");
  Ut.assert_bool n "binOr"  (binOr  "1100" "1010" = "1110");
;;





