(*
    Confluence System Design Language Compiler
    Copyright (C) 2003-2004 Tom Hawkins (tomahawkins@yahoo.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*)


open Nls;;


(* Misc helpers. *)

let infoIdSet ids =
  "{" ^ IdSet.fold (fun id str -> str ^ " " ^ string_of_int (idToInt id)) ids "" ^ " }";;

let spaceInfo space levels =

  let rec f indent space levels =
    indent ^ "Root      : " ^ (if space.spaceIsRoot then "Yes" else "No") ^ "\n" ^
    indent ^ "Inputs    : " ^ string_of_int (IdSet.cardinal space.spaceInputs)    ^ " : " ^ infoIdSet space.spaceInputs ^ "\n" ^
    indent ^ "Outputs   : " ^ string_of_int (IdSet.cardinal space.spaceOutputs)   ^ " : " ^ infoIdSet space.spaceOutputs ^ "\n" ^
    indent ^ "Prims     : " ^ string_of_int (IdSet.cardinal space.spacePrims)     ^ " : " ^ infoIdSet space.spacePrims ^ "\n" ^
    indent ^ "Producers : " ^ string_of_int (IdSet.cardinal space.spaceProducers) ^ (*" : " ^ infoIdList (IdSet.elements space.spaceProducers) ^*) "\n" ^
    indent ^ "Consumers : " ^ string_of_int (IdSet.cardinal space.spaceConsumers) ^ (*" : " ^ infoIdList (IdSet.elements space.spaceConsumers) ^*) "\n" ^
    indent ^ "Subs      : " ^ string_of_int (List.length space.spaceSubs) ^ "\n" ^
    List.fold_left (fun str sub -> str ^ indent ^ "  Sub:" ^ (if levels <= 0 then " Not Displayed\n" else "\n" ^ f (indent ^ "    ") sub (levels - 1))) "" space.spaceSubs
  in
  "SpaceInfo:\n" ^ (f "  " space levels);;


(** Populate nodes in spaces. *)

let rec connectChild child =
  if not child.spaceIsRoot && not (List.memq child child.spaceParent.spaceSubs) then begin
    child.spaceParent.spaceSubs <- child :: child.spaceParent.spaceSubs;
    connectChild child.spaceParent
  end
;;

let rec disconnectChild child =
  if not child.spaceIsRoot then begin
    child.spaceParent.spaceSubs <- List.filter (fun space -> space != child) child.spaceParent.spaceSubs;
    if child.spaceParent.spaceSubs == [] && IdSet.is_empty child.spaceParent.spacePrims then disconnectChild child.spaceParent;
  end
;;

(** Populates spaces with primitive and subspace information. *)
let populateSpaces nls =
  Hashtbl.iter (fun id space ->
    connectChild space;  (* XXX Is this correct with Lp.isCoded?  Conflicts with disconnectChild. *)
    if Lp.isCoded (node nls id) then space.spacePrims <- IdSet.add id space.spacePrims
  ) nls.n2s
;;


(** Generates XML netlist. *)

let write = NlsCode.write;;

let codeXmlNetlist nls =
  let fileName = getFile nls in
  let buildName = getBuild nls in
  let docs = getHeader nls in
  let src id port =
    "n" ^ string_of_int (idToInt (source nls id port))
  in
  let intList i ints =
    List.iter (fun int -> write (i ^ "<integer value=\"" ^ string_of_int int ^ "\"/>\n")) ints
  in
  let intbigList i ints =
    List.iter (fun int -> write (i ^ "<integer value=\"" ^ Intbig.string_of_intbig int ^ "\"/>\n")) ints
  in
  let prim1 i prim id fields =
    write (i ^ "<prim_" ^ prim ^ " id=\"n" ^ string_of_int (idToInt id) ^ "\"");
    List.iter (function (key, value) -> write (" " ^ key ^ "=\"" ^ String2.replace_xml_chars value ^ "\"")) fields;
    write (">\n")
  in
  let prim0 i prim id fields =
    write (i ^ "<prim_" ^ prim ^ " id=\"n" ^ string_of_int (idToInt id) ^ "\"");
    List.iter (function (key, value) -> write (" " ^ key ^ "=\"" ^ String2.replace_xml_chars value ^ "\"")) fields;
    write ("/>\n")
  in
  let prim i id =
    match node nls id with
      Lp.VecInput   (n, p, w) -> prim0 i "input"  id [("name", n); ("position", string_of_int p); ("width", string_of_int w)]
    | Lp.VecOutput  (n, p, w) -> prim0 i "output" id [("name", n); ("position", string_of_int p); ("width", string_of_int w); ("output", src id 0)]
    | Lp.VecInOut   (n, p, w) -> prim0 i "inout"  id [("name", n); ("position", string_of_int p); ("width", string_of_int w); ("ctrl", src id 0); ("output", src id 1)]
    | Lp.VecAssert  (m)       -> prim0 i "assert" id [("message", m); ("input", src id 0)]
    | Lp.VecPrint   (m)       -> prim0 i "print"  id [("enable", src id 0); ("input", src id 1)]
    | Lp.VecConst   (v)       -> prim0 i "const"  id [("value", v)]
    | Lp.VecBuf     (w)       -> prim0 i "buf"    id [("width", string_of_int w); ("input", src id 0)]
    | Lp.VecNot     (w)       -> prim0 i "not"    id [("width", string_of_int w); ("input", src id 0)]
    | Lp.VecAnd     (w)       -> prim0 i "and"    id [("width", string_of_int w); ("left", src id 0); ("right", src id 1)]
    | Lp.VecOr      (w)       -> prim0 i "or"     id [("width", string_of_int w); ("left", src id 0); ("right", src id 1)]
    | Lp.VecXor     (w)       -> prim0 i "xor"    id [("width", string_of_int w); ("left", src id 0); ("right", src id 1)]
    | Lp.VecConcat  (l, r)    -> prim0 i "concat" id [("left_width", string_of_int l); ("right_width", string_of_int r); ("left", src id 0); ("right", src id 1)]
    | Lp.VecSelect  (w, bs) ->
        prim1 i "select" id [("width", string_of_int w); ("input", src id 0)];
        intList (i ^ " ") bs;
        write (i ^ "</prim_select>\n")
    | Lp.VecEqu     (w)       -> prim0 i "equ"    id [("width", string_of_int w); ("left", src id 0); ("right", src id 1)]
    | Lp.VecAdd     (w)       -> prim0 i "add"    id [("width", string_of_int w); ("left", src id 0); ("right", src id 1)]
    | Lp.VecSub     (w)       -> prim0 i "sub"    id [("width", string_of_int w); ("left", src id 0); ("right", src id 1)]
    | Lp.VecMul     (w)       -> prim0 i "mul"    id [("width", string_of_int w); ("left", src id 0); ("right", src id 1)]
    | Lp.VecSwitch  (w)       -> prim0 i "switch" id [("width", string_of_int w); ("predicate", src id 0); ("true", src id 1); ("false", src id 2)]
    | Lp.VecExtComb (n, ps, iw, o) ->
        prim1 i "extcomb" id [("name", n); ("input_width", string_of_int iw); ("output_width", string_of_int o); ("input", src id 0)];
        intList (i ^ " ") ps;
        write (i ^ "</prim_extcomb>\n")
    | Lp.VecExtSequ (c, n, ps, iw, o) ->
        prim1 i "extsequ" id [("clock", c); ("name", n); ("input_width", string_of_int iw); ("output_width", string_of_int o); ("reset", src id 0); ("enable", src id 1); ("input", src id 2)];
        intList (i ^ " ") ps;
        write (i ^ "</prim_extsequ>\n")
    | Lp.VecReg     (c, w)     -> prim0 i "reg"    id [("clock", c); ("width", string_of_int w); ("reset", src id 0); ("enable", src id 1); ("input", src id 2)]
    | Lp.VecRom     (c, w, vs) ->
        prim1 i "rom" id [("clock", c); ("width", string_of_int w); ("enable", src id 0); ("addr", src id 1)];
        intbigList (i ^ " ") vs;
        write (i ^ "</prim_rom>\n")
    | Lp.VecRamMem  (dw, vs) ->
        prim1 i "rammem" id [("width", string_of_int dw)];
        intbigList (i ^ "  ") vs;
        write (i ^ "</prim_rammem>\n")
    | Lp.VecRamWrite (c, aw, dw)      -> prim0 i "ramwrite" id [("clock", c); ("addr_width", string_of_int aw); ("data_width", string_of_int dw); ("memory", src id 0); ("enable", src id 1); ("addr", src id 2); ("data", src id 3)]
    | Lp.VecRamRead  (c, rbw, aw, dw) -> prim0 i "ramread"  id [("clock", c); ("is_rbw", (if rbw then "true" else "false")); ("addr_width", string_of_int aw); ("data_width", string_of_int dw); ("memory", src id 0); ("enable", src id 1); ("addr", src id 2)]
  in
  let rec hier i space =
    write (i ^ "<hierarchy instance=\"" ^ String2.replace_xml_chars space.spaceInstLoc ^
      "\" component=\"" ^ String2.replace_xml_chars space.spaceCompLoc ^ "\">\n");
    IdSet.iter (fun id -> prim (i ^ " ") id) space.spacePrims;
    List.iter (fun subspace -> hier (i ^ " ") subspace) space.spaceSubs;
    write (i ^ "</hierarchy>\n");
  in
  NlsCode.openFile (fileName ^ ".xml");
  write ("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n");
  write ("<!DOCTYPE netlist SYSTEM \"http://www.launchbird.com/misc/confluence_netlist.dtd\">\n");
  write ("<netlist name=\"" ^ String2.replace_xml_chars buildName ^ "\">\n");
  write (" <documentation>\n");
  write (String2.replace_xml_chars docs ^ "\n");
  write (" </documentation>\n");
  write (" <interface>\n");
  List.iter (fun clock -> write ("  <clock name=\"" ^ String2.replace_xml_chars clock ^ "\"/>\n")) (getClocks nls);
  List.iter (fun pid -> match node nls pid with
      Lp.VecInput  (n, p, w) -> write ("  <input name=\"" ^ String2.replace_xml_chars n ^ "\" position=\"" ^ string_of_int p ^ "\" width=\"" ^ string_of_int w ^ "\"/>\n");
    | Lp.VecOutput (n, p, w) -> write ("  <output name=\"" ^ String2.replace_xml_chars n ^ "\" position=\"" ^ string_of_int p ^ "\" width=\"" ^ string_of_int w ^ "\"/>\n");
    | Lp.VecInOut  (n, p, w) -> write ("  <inout name=\"" ^ String2.replace_xml_chars n ^ "\" position=\"" ^ string_of_int p ^ "\" width=\"" ^ string_of_int w ^ "\"/>\n");
    | _ -> raise (Error "NlsSpace.codeXmlNetlist")
  ) (portList nls);
  write (" </interface>\n");
  hier " " nls.rootSpace;
  write ("</netlist>\n");
  NlsCode.closeFile ()
;;


(** Rearrange (ports and ram_ports) and condense spaces. *)

(** Signal map generation. *)
let codeSignalMap nls =
  let file = open_out (getFile nls ^ "_signal_map.txt") in
  let write str =
    output_string file str
  in
  let rec f0 locs =
    match locs with
    | [] -> ""
    | loc :: locs -> f0 locs ^ "  " ^ Loc.toString loc ^ "\n"
  in
  List.iter (fun node -> if Lp.isCoded (Nls.node nls node) then write (NlsCode.signalName nls node ^ "(" ^ Lp.primString (Nls.node nls node) ^ "):\n" ^ f0 (Nls.nodeLocs nls node) ^ "\n")) (Nls.nodeList nls);
  close_out file
;;


