(*
    Confluence System Design Language Compiler
    Copyright (C) 2003-2004 Tom Hawkins (tomahawkins@yahoo.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*)


open Nls;;

let write = NlsCode.write;;

Lp.setWriter write;;


let codeSignalDecl name width =
  if width = 1 then
    name
  else
    "[" ^ string_of_int (width - 1) ^ ":0] " ^ name;;

let codePort nls id =
  match node nls id with
    Lp.VecInput    (name, _, w) -> "input  " ^ codeSignalDecl (name ^ "_i") w ^ ";\n"
  | Lp.VecOutput   (name, _, w) -> "output " ^ codeSignalDecl (name ^ "_o") w ^ ";\n"
  | Lp.VecInOut    (name, _, w) -> "inout  " ^ codeSignalDecl (name ^ "_b") w ^ ";\n"
  | _ -> raise (Error "codePort: Expecting port node.");;

let formatPortNames nls =
  List.map (function clock -> clock ^ "_c") (getClocks nls) @
  List.map (function id ->
    match node nls id with
      Lp.VecInput    (name, _, _) -> name ^ "_i"
    | Lp.VecOutput   (name, _, _) -> name ^ "_o"
    | Lp.VecInOut    (name, _, _) -> name ^ "_b"
    | _ -> raise (Error "formatPortNames: Expecting port node."))
    (portList nls);;


(* Module generator. *)

let code nls =
  let nodes = nodeList nls in
  let buildName = getBuild nls in
  NlsCode.openFile (getFile nls ^ ".v");
  NlsCode.codeDocs nls "//" (getHeader nls) buildName;
  write ("`timescale 1 ns / 1 ns\n\n");
  write ("module " ^ buildName ^ " "); Lp.argList (formatPortNames nls); write ";\n";
  List.iter (fun clock -> write ("input  " ^ clock ^ "_c;\n")) (getClocks nls);
  List.iter (fun node -> LpVerilog.decl (Nls.node nls node) (NlsCode.signalName nls node)) nodes;
  List.iter (fun node -> LpVerilog.func (Nls.node nls node) (NlsCode.signalName nls node) (List.map (fun input -> NlsCode.signalName nls input) (Nls.sourceList nls node))) nodes;
  write "endmodule\n\n";
  NlsCode.closeFile ()
;;


(* Testbench generator. *)

let rec codePortArgs nls ids =
  match ids with
    [] -> ""
  | [id] -> NlsCode.signalName nls id
  | id :: ids -> NlsCode.signalName nls id ^ ", " ^ codePortArgs nls ids
;;

let codeTestBench nls cycles =
  let fileName = getFile nls in
  let buildName = getBuild nls in
  let docs = getHeader nls in
  let clocks = Nls.getClocks nls in
  let ports = Nls.portList nls in
  if List.length clocks <> 1 then raise (Error "TestBench generation requires single clock design.");
  let clock = List.hd clocks in
  NlsCode.openFile (fileName ^ "_bench.v");
  NlsCode.codeDocs nls "//" docs buildName;
  write ("`timescale 1 ns / 1 ns\n\n");
  write ("module " ^ buildName ^ "_bench;\n\n");
  write ("integer cycle_count;\n");
  write ("reg " ^ clock ^ "_c;\n");
  List.iter (function id -> write ("wire " ^ codeSignalDecl (NlsCode.signalName nls id) (nodeWidth nls id) ^ ";\n")) ports;
  write ("\ninitial begin\n");
  write ("  " ^ clock ^ "_c = 1'b0;\n");
  write ("  $dumpfile(\"" ^ fileName ^ "_v.vcd\");\n");
  List.iter (function id -> write ("  $dumpvars(1, " ^ buildName ^ "_bench." ^ NlsCode.signalName nls id ^ ");\n")) ports;
  write ("  for (cycle_count = " ^ string_of_int cycles ^ "; cycle_count > 0; cycle_count = cycle_count - 1) begin\n");
  write ("    #1 " ^ clock ^ "_c = 1'b1;\n");
  write ("       " ^ clock ^ "_c = 1'b0;\n");
  write ("  end\n");
  write ("  #1 $finish;\n");
  write ("end\n\n");
  write (buildName ^ " " ^ buildName ^ "(" ^ clock ^ "_c, " ^ codePortArgs nls ports ^ ");\n\n");
  write ("endmodule\n\n");
  NlsCode.closeFile ()
;;

