(*
    Confluence System Design Language Compiler
    Copyright (C) 2003-2004 Tom Hawkins (tomahawkins@yahoo.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*)


open Nls;;

let write = NlsCode.write;;

Lp.setWriter write;;

let codeUnsigned width =
  "unsigned(" ^ string_of_int (width - 1) ^ " downto 0)";;

let rec codeClockList clocks =
  match clocks with
    [] -> ""
  | clock :: clocks -> "signal " ^ clock ^ "_c : in std_logic;\n" ^ codeClockList clocks;;

let codePortList nls =
  let f sofar pid =
    match node nls pid with
      Lp.VecInput (name, _, w) -> sofar ^ "signal " ^ name ^ "_i : in " ^ codeUnsigned w ^ ";\n"
    | Lp.VecOutput(name, _, w) -> sofar ^ "signal " ^ name ^ "_o : out " ^ codeUnsigned w ^ ";\n"
    | Lp.VecInOut (name, _, w) -> sofar ^ "signal " ^ name ^ "_b : inout " ^ codeUnsigned w ^ ";\n"
    | _ -> raise (Error "codePortList: Expecting port node.")
  in
    let s = codeClockList (getClocks nls) ^ List.fold_left f "" (portList nls)
    in
      if s = "" then "" else String.sub s 0 (String.length s - 2);;

let rec codeExtGeneric pNum pTotal =
  if pNum < pTotal then begin
    write ("; p" ^ string_of_int pNum ^ " : integer");
    codeExtGeneric (pNum + 1) pTotal
  end;;

let codeExtDecl lp =
  match lp with
    Lp.VecExtComb (name, params, wi, wo) ->
      write ("component " ^ name ^ " is\n");
      write ("generic (input_width : integer; output_width : integer");
      codeExtGeneric 0 (List.length params);
      write (");\n");
      write ("port (i : in unsigned(input_width - 1 downto 0); o : out unsigned(output_width - 1 downto 0));\n");
      write ("end component " ^ name ^ ";\n")
  | Lp.VecExtSequ (cd, name, params, wi, wo) ->
      write ("component " ^ name ^ " is\n");
      write ("generic (input_width : integer; output_width : integer");
      codeExtGeneric 0 (List.length params);
      write (");\n");
      write ("port (clock : in std_logic; reset : in unsigned(0 downto 0); enable : in unsigned(0 downto 0); i : in unsigned(input_width - 1 downto 0); o : out unsigned(output_width - 1 downto 0));\n");
      write ("end component " ^ name ^ ";\n")
  | _ -> raise (Failure "NlsVhdl.codeExtDecl: Should not get here.")
;;





(* Module (entity/arch pair) generator. *)

let rec uniqueList items sofar =
  match items with
  | [] -> sofar
  | item :: items -> uniqueList items (if List.mem item sofar then sofar else item :: sofar)
;;

let code nls =
  let nodes = nodeList nls in
  let externalLps = List.fold_left (fun sofar node -> let node = Nls.node nls node in if (match node with Lp.VecExtComb _ | Lp.VecExtSequ _ -> true | _ -> false) && not (List.mem node sofar) then node :: sofar else sofar) []  nodes in
  let buildName = getBuild nls in
  NlsCode.openFile (getFile nls ^ ".vhd");
  NlsCode.codeDocs nls "--" (getHeader nls) buildName;
  write "library ieee;\nuse ieee.std_logic_1164.all;\nuse ieee.numeric_std.all;\n";
  write ("entity " ^ buildName ^ " is\n");
  write ("port(\n");
  write (codePortList nls);
  write (");\nend entity " ^ buildName ^ ";\n");
  write ("architecture rtl of " ^ buildName ^ " is\n");
  List.iter (fun node -> LpVhdl.decl (Nls.node nls node) (NlsCode.signalName nls node)) nodes;
  List.iter codeExtDecl externalLps;
  write "begin\n";
  List.iter (fun node -> LpVhdl.func (Nls.node nls node) (NlsCode.signalName nls node) (List.map (fun input -> NlsCode.signalName nls input) (Nls.sourceList nls node))) nodes;
  write "end architecture rtl;\n\n\n";
  NlsCode.closeFile ()
;;


(* Testbench generator. *)

let rec codePortArgs nls ids =
  match ids with
    [] -> ""
  | id :: ids -> ", " ^ NlsCode.signalName nls id ^ codePortArgs nls ids;;

let codeTestBench nls cycles =
  let fileName = getFile nls in
  let buildName = getBuild nls in
  let docs = getHeader nls in
  let clocks = Nls.getClocks nls in
  let ports = Nls.portList nls in
  if List.length clocks <> 1 then raise (Error "TestBench generation requires single clock design.");
  let clock = List.hd clocks in
  NlsCode.openFile (fileName ^ "_bench.vhd");
  NlsCode.codeDocs nls "--" docs buildName;
  write "library ieee;\nuse ieee.std_logic_1164.all;\nuse ieee.numeric_std.all;\n";
  write ("entity " ^ buildName ^ "_bench is\n");
  write ("end entity " ^ buildName ^ "_bench;\n");
  write ("architecture rtl of " ^ buildName ^ "_bench is\n");
  write ("component " ^ buildName ^ " is\n");
  write ("port(\n");
  write (codePortList nls);
  write (");\nend component " ^ buildName ^ ";\n");
  write ("signal " ^ clock ^ "_c : std_logic;\n");
  List.iter (fun id -> write ("signal " ^ NlsCode.signalName nls id ^ " : " ^ codeUnsigned (nodeWidth nls id) ^ ";\n")) ports;
  write ("begin\n");
  write ("process\n");
  write ("variable cycle_count : integer := " ^ string_of_int cycles ^ ";\n");
  write ("begin\n");
  write ("  while cycle_count > 0 loop\n");
  write ("    cycle_count := cycle_count - 1;\n");
  write ("    wait for 1 ns;\n");
  write ("    " ^ clock ^ "_c <= '1';\n");
  write ("    wait until " ^ clock ^ "_c = '1';\n");
  write ("    " ^ clock ^ "_c <= '0';\n");
  write ("  end loop;\n");
  write ("end process;\n");
  write ("s1 : " ^ buildName ^ " port map (" ^ clock ^ "_c" ^ codePortArgs nls ports ^ ");\n");
  write "end architecture rtl;\n\n";
  NlsCode.closeFile ()
;;

